#include <M5Stack.h>
#include <TimeLib.h>                              //By Michael Margolis
#include <WiFi.h>
#include <WiFiUdp.h>
#include "EEPROM.h"

/*
   Copyright Sergio Daniel CastaÃ±eda NiÃ±o 2018
   PLC Arduino Firmware M5Stack V1.0
*/

#define BYTE6MASK0 0x3F
#define BYTE6MASK1 0xFC
#define BYTE6MASK2 0X3F000
#define BYTE6MASK3 0xFC0000
#define BYTE6MASK4 0x3F000000
#define BYTE6MASK5 0xC0000000

#define MDMASK 0XFC0
#define HDMASK 0xF000

#define BYTEMASK0 0xFF
#define BYTEMASK1 0xFF00
#define BYTEMASK2 0xFF0000
#define BYTEMASK3 0xFF000000

#define GMASK8 0xFF
#define GMASK16 0xFFFF

#define REG_INT 2866
#define REG_LONG 2931
#define REG_FLOAT 3156
#define REG_TIMER 3285
#define REG_COUNTER 3574
#define REG_STATUS 3959
#define REG_IN 3960
#define REG_OUT 3976
#define REG_VIRTUAL 3992
#define REG_TYPE_SIZE 4057

#define LEDC_BASE_FREQ_PWM     5000
#define LEDC_BASE_SERVO     50

#define PINO0X 0
#define PINO0Y 0
#define PINO1X 80
#define PINO1Y 0
#define PINO2X 160 
#define PINO2Y 0
#define PINO3X 240
#define PINO3Y 0

#define PINI0X 13
#define PINI0Y 160
#define PINI1X 121
#define PINI1Y 160
#define PINI2X 228
#define PINI2Y 160

#define RECT_W 72
#define RECT_H 54

#define LIGHT_GREEN 0x07E0
#define DARK_GREEN 0x0440
#define LIGHT_RED 0xF800
#define DARK_RED 0x7800
#define GRAY_BACK 0x6331
#define YELLOW 0xFFE0
#define BLUE_C 0x4D3F
#define ORANGE 0xFBE4


const char* ssid = "";
const char* password = "";  


/*
   HMI MODE
   HMI mode allows you to connect your board using this firmware with the HMI Controller for Arduino aplication for Android.

   Press the A button to enable the HMI mode in the M5Stack.

*/


int clockGMT = -500;

/* SET TIME ZONE GMT (GREENWICH MEAN TIME)
   Set your time zone GMT in the "clockGMT" variable.
   Example 1: if your time zone is GMT+2:00 set clockGMT = 200;
   Example 2: if your time zone is GMT-5:00 set clockGMT = -500;
   Example 3: if your time zonie is GMT+9:30 set clockGMT = 930;

   Name   Description                         Relative to GMT
   GMT    Greenwich Mean Time                       GMT
   UTC    Universal Coordinated Time                GMT
   ECT    European Central Time                     GMT+1:00
   EET    Eastern European Time                     GMT+2:00
   ART    (Arabic) Egypt Standard Time              GMT+2:00
   EAT    Eastern African Time                      GMT+3:00
   MET    Middle East Time                          GMT+3:30
   NET    Near East Time                            GMT+4:00
   PLT    Pakistan Lahore Time                      GMT+5:00
   IST    India Standard Time                       GMT+5:30
   BST    Bangladesh Standard Time                  GMT+6:00
   VST    Vietnam Standard Time                     GMT+7:00
   CTT    China Taiwan Time                         GMT+8:00
   JST    Japan Standard Time                       GMT+9:00
   ACT    Australia Central Time                    GMT+9:30
   AET    Australia Eastern Time                    GMT+10:00
   SST    Solomon Standard Time                     GMT+11:00
   NST    New Zealand Standard Time                 GMT+12:00
   MIT    Midway Islands Time                       GMT-11:00
   HST    Hawaii Standard Time                      GMT-10:00
   AST    Alaska Standard Time                      GMT-9:00
   PST    Pacific Standard Time                     GMT-8:00
   PNT    Phoenix Standard Time                     GMT-7:00
   MST    Mountain Standard Time                    GMT-7:00
   CST    Central Standard Time                     GMT-6:00
   EST    Eastern Standard Time                     GMT-5:00
   IET    Indiana Eastern Standard Time             GMT-5:00
   PRT    Puerto Rico and US Virgin Islands Time    GMT-4:00
   CNT    Canada Newfoundland Time                  GMT-3:30
   AGT    Argentina Standard Time                   GMT-3:00
   BET    Brazil Eastern Time                       GMT-3:00
   CAT    Central African Time                      GMT-1:00
*/



unsigned int udpLocalPort = 8888;
char timeServer[] = "time.nist.gov"; // time.nist.gov NTP server
const int NTP_PACKET_SIZE = 48; // NTP time stamp is in the first 48 bytes of the message
byte packetBuffer[ NTP_PACKET_SIZE]; //buffer to hold incoming and outgoing packets


WiFiUDP Udp;
unsigned long epoch;
unsigned long currentGetRTCUpdateTime;
const int targetGetRTCUpdateTime = 300;

unsigned long startReprogramTime;
const unsigned long targetReprogramTime = 15000;


bool inReprogramTime = true;
bool runReprogramTime = false;

bool wifiConnected = false;
bool clientConnected = false;
bool rtcUpdated = false;
bool updateMyTime = false;

int countPC = 0;
int countCounter = 0;
int countCounterCTU = 0;
int countCounterCTUD = 0;
int countTimer = 0;
int counterInt = 0;
int counterDint = 0;
int counterReal = 0;

char* namePinO0 = "O.0";
char* namePinO1 = "O.1";
char* namePinO2 = "O.2";
char* namePinO3 = "O.3";

char* namePinI0 = "I.0";
char* namePinI1 = "I.1";
char* namePinI2 = "I.2";

bool runningProgram = false;
bool displayState = true;
bool programming = false;

unsigned long currentTimeDisplay;
unsigned long startTimeDisplay;

bool preIState[3] = {true, true, true};
bool preOState[4] = {true, true, true, true};


bool iState[3] = {false, false, false};
bool oState[4] = {false, false, false, false};
bool vState[64] = {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,
                   false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false
                  };

bool iPreState[3] = {false, false, false};
bool oPreState[4] = {false, false, false, false};
bool vPreState[64] = {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,
                      false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,
                     };

bool iMask[3] = {false, false, false};
bool oMask[4] = {false, false, false, false};
bool vMask[64] = {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,
                  false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false
                 };

bool iUpEdge[3] = {false, false, false};
bool oUpEdge[4] = {false, false, false, false};
bool vUpEdge[64] = {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,
                    false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false
                   };

bool iDownEdge[3] = {false, false, false};
bool oDownEdge[4] = {false, false, false, false};
bool vDownEdge[64] = {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,
                      false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false
                     };

int16_t intValues[32] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int32_t dintValues[56] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
float realValues[32]  = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

int8_t iIndex[3] = {0, 0, 0};
int8_t oIndex[4] = {0, 0, 0, 0};
int8_t vIndex[64] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int8_t intIndex[32] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int8_t dintIndex[56] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int8_t realIndex[32] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

uint8_t iPins[3] = {26, 35, 36};    //INPUT pins
uint8_t oPins[4] = {16, 17, 2, 5};      //OUTPUT pins

bool analogIEnabled[3] = {false, false, false};
int analogInputs[3] = {0,0,0};
int preAnalogInputs[3] = {1,1,1};

int preAnalogOutputs[4] = {1,1,1,1};
int analogOutputs[4] = {0, 0, 0, 0};                            //change this
bool pwmOutput[4] = {false, false, false, false};
bool servoOutput[4] = {false, false, false, false};

int16_t timeData[7] = {0, 0, 0, 0, 0, 0, 0};


typedef union {
  float val;
  uint8_t bytes[4];
} floatValue;

struct Counter {
  uint8_t no;
  uint8_t reset;
  bool dn;
  bool dnNull;
  bool countNo;
  bool resetNo;
  bool countUp;
  bool countUpNull;
  uint8_t noIndex;
  uint8_t resetIndex;
  uint8_t dnIndex;
  uint8_t countUpIndex;
  uint8_t presetIndex;
  uint8_t countIndex;
};

struct CounterCTUD {
  uint8_t no;
  uint8_t noDown;
  uint8_t reset;
  bool dn;
  bool dnNull;
  bool countNo;
  bool countNoDown;
  bool resetNo;
  bool countUp;
  bool countDown;
  bool countUpNull;
  bool countDownNull;
  uint8_t noIndex;
  uint8_t noDownIndex;
  uint8_t resetIndex;
  uint8_t dnIndex;
  uint8_t countUpIndex;
  uint8_t countDownIndex;
  uint8_t presetIndex;
  uint8_t countIndex;
};


struct Timer {
  uint8_t no;
  bool dn;
  bool dnNull;
  uint8_t reset;
  bool tt;
  bool enableNo;
  bool activeTimer;
  bool ttNull;
  bool timerDone;
  int typeTimer;
  uint8_t noIndex;
  uint8_t dnIndex;
  uint8_t ttIndex;
  uint8_t resetIndex;
  int8_t presetIndex;
  int8_t currentTimeIndex;
  unsigned long startTime;
  bool isPoint1;
};


unsigned long startTime;
unsigned long delta;
unsigned long targetTime;

unsigned long PC; //HERE
uint8_t n;
int mask = (0x3FFF);
bool A;
bool X;
bool Y;

bool notFlag;
bool risingFlag;
bool fallingFlag;
bool iFlag;
bool iFlag1;
bool oFlag1;
bool oFlag;
bool vFlag;
bool vFlag1;
bool vFlag2;
bool vFlag3;
bool vFlag4;
bool vFlag5;
bool vFlag6;

bool latchFlag;
bool unlatchFlag;
bool negateFlag;
bool coilFlag;

bool timerFlag;
bool counterFlag;
bool counterCTUDFlag;

bool inFunction;

bool configStream = false;
bool enableHMIComm = false;
int8_t inputSize;
int8_t outputSize;
int8_t virtualSize;

struct Counter myCounter[16];
struct CounterCTUD myCounterCTUD[16];
struct Timer myTimer[16];
char bufferSerial[128];
int buffSize;
bool inCount;
int8_t waitReadCounter;
uint8_t ld;
uint8_t md;
uint8_t hd;
uint8_t ld32;
uint8_t md32;
uint8_t hd32;
uint16_t tempValue;
uint32_t tempValue32;
bool sendDataConfig = false;
bool sendStartSend = false;

bool alreadySendIt = false;
bool inProcessLock = true;
bool requestSecondPart = false;
bool inSecondState = false;

int8_t iSize, oSize, vSize, eSize, dSize, floatSize;

unsigned long startTimeSend;
unsigned long targetTimeSend = 1000;

uint8_t preEEPROM[4096];

WiFiServer server(23);
WiFiClient clientTemp;

void checkSpace();
void reprogram();
void initProgram();
void interpreter();
void updateRegisters();
bool setDnCounter(int posi, bool dnState);
bool setDnCounterCTUD(int posi, bool dnState);
void cleanOldCounters(int tempCountCounter);
void cleanOldTimers(int tempCountTimer);
void checkVariablesDint();
int readAnalogPin(int index);
void writeAnalogPin(int index, int value);
int readAnalogOutput(int index);
void setDnValue(bool newState, int tempPC);
bool getNoValue();
float getSource();
void setDest(float valueDest);
int16_t getJoinValue();
bool isTagEnable(int tempPC);

void checkVariablesAOut();
void checkVariablesInt();
void checkVariablesDint();
void checkVariablesReal();
void readRawVariables();
bool isSourceInt();
int getOutIndex();
bool getUpEdgeActive();
void updateVariable(int tempPC, int32_t valueDest);
void setDest(int tempPC, float valueDest);

void equ();
void neq();
void les();
void leq();
void grt();
void geq();
void mov();
void lim();
void meq();
void andOp();
void orOp();
void notOp();
void xorOp();
void addOp();
void subOp();
void mulOp();
void divOp();
void xpy();
void absMath();
void sqrMath();
void asnMath();
void acsMath();
void atnMath();
void cosMath();
void lnMath();
void logMath();
void sinMath();
void tanMath();
void scp();
void scl();

void bsl();
void bsr();
void ror();
void rol(); 



void task1(void * pvParameters) {

    for(;;) {
      if(runningProgram){
        updateUI();

        if(inReprogramTime)
          drawPLCWord();
        else
          drawHMIWord();
          
        drawCurrentTime();

        if(programming){
          showProgramming(); 
          showResetM5();
        }
         
      }
      else{
        if(programming){
          showProgramming();
          showResetM5();
        }
      }//END else
      
      delay(200);
    }
}



void setup() {
  for (int i = 0; i < 3; i++)   //pinMode for all the pins of the PLC
    pinMode(iPins[i], INPUT);
  for(int i=0; i<4; i++)
    pinMode(oPins[i], OUTPUT);
  
  initADC(); 
  initAnalogOut();

  Serial.begin(115200);

  M5.begin();

  xTaskCreatePinnedToCore(task1,"task1",4096,NULL,1,NULL,0);      

  M5.Lcd.println();
  M5.Lcd.println(" PLC Ladder Simulator for M5Stack");
  M5.Lcd.println();

  M5.Lcd.setTextSize(1);

  startWifi();

  EEPROM.begin(4096);


  updateTimeRegisters();

  if(EEPROM.read(3959) != '2'){
    if(wifiConnected){
      M5.Lcd.println();
      M5.Lcd.println(" No ladder program in memory");
      M5.Lcd.println();
      M5.Lcd.setTextColor(RED, BLACK);
      M5.Lcd.println(" UPLOAD LADDER PROGRAM");
    }
  }

  while (EEPROM.read(3959) != '2') {
    if (wifiConnected)
      reprogram();
   
    delay(400);
  }

  runReprogramTime = false;

  initProgram();

}

void loop() {
  // put your main code here, to run repeatedly:
  if (EEPROM.read(3959) == '2') {
    runningProgram = true;
    drawBackground();
    preDraw();
    

    while (true) {
      interpreter();
      updateRegisters();
      updateTimeRegisters();

      if (wifiConnected)
        reprogram();

      checkInputBtn();

      M5.update();
      delay(20);
     
    }//END while

  }//END IF EEPROM.read
}//END loop()


void interpreter() {
  PC = 0; //put the PC in the start of the memory
  int dataByte = EEPROM.read(PC);

  while (dataByte != 'z') { //z is the value that indicate that the program has ended

    if (inFunction) {
      switch (dataByte) {
        case 'a':
          equ();
          break;
        case 'b':
          neq();
          break;
        case 'c':
          les();
          break;
        case 'd':
          leq();
          break;
        case 'e':
          grt();
          break;
        case 'f':
          geq();
          break;
        case 'g':
          lim();
          break;
        case 'h':
          meq();
          break;
        case 'i':
          mov();
          break;
        case 'j':
          bsl();
          break;
        case 'k':
          bsr();
          break;
        case 'l':
          rol();
          break;
        case 'm':
          ror();
          break;
        case 'n':
          andOp();
          break;
        case 'o':
          orOp();
          break;
        case 'p':
          notOp();
          break;
        case 'q':
          xorOp();
          break;
        case 'r':
          addOp();
          break;
        case 's':
          subOp();
          break;
        case 't':
          mulOp();
          break;
        case 'u':
          divOp();
          break;
        case 'v':
          absMath();
          break;
        case 'w':
          sqrMath();
          break;
        case 'x':
          xpy();
          break;
        case 'y':
          asnMath();
          break;
        case 'I':
          acsMath();
          break;
        case 'A':
          atnMath();
          break;
        case 'B':
          cosMath();
          break;
        case 'C':
          lnMath();
          break;
        case 'D':
          logMath();
          break;
        case 'E':
          sinMath();
          break;
        case 'F':
          tanMath();
          break;
        case 'G':
          scp();
          break;
        case 'H':
          scl();
          break;
        default:
          Serial.println("ERROR");
          break;

      }//END swith()
      inFunction = false;
    }

    else {
      switch (dataByte) {
        case '$': //functions
          inFunction = true;
          break;
        case 'I': //i
          iFlag = true;
          break;
        case 'i':
          iFlag1 = true;
          break;
        case 'O': //o
          oFlag = true;
          break;
        case 'k':
          oFlag1 = true;
          break;
        case 'V': //v
          vFlag = true;
          break;
        case'v':
          vFlag1 = true;
          break;
        case 'U': //v
          vFlag2 = true;
          break;
        case'u':
          vFlag3 = true;
          break;
        case 'A':
          vFlag4 = true;
          break;
        case 'a':
          vFlag5 = true;
          break;
        case 'c':
          vFlag6 = true;
          break;
        case 'L': //Latch
          latchFlag = true;
          break;
        case 'l': //Unlatch
          unlatchFlag = true;
          break;
        case '|': //or
          Y = Y | A;
          A = true;
          break;
        case '}': //and
          A = Y | A;
          Y = false;
          break;
        case '~': //not
          notFlag = true;
          break;
        case 'e': //RISING EDGE
          risingFlag = true;
          break;
        case 'f': //FALLING EDGE
          fallingFlag = true;
          break;
        case 'C': //coil
          coilFlag = true;
          break;
        case 'n': //negate coil
          negateFlag = true;
          break;
        case 'R': //rung
          A = true;
          Y = false;
          coilFlag = false;
          break;
        case 'q': //timer
          timerFlag = true;
          coilFlag = false;
          break;
        case 'B': //counter CTU
          counterFlag = true;
          coilFlag = false;
          break;
        case 'b': //counter CTUD
          counterCTUDFlag = true;
          coilFlag = false;
          break;
        default:
          dataByte = dataByte - 48;
          if ((dataByte < 15) && (dataByte >= 0)) { //posible to be changed

            n = dataByte;

            if (coilFlag) {
              if (iFlag) {
                iState[n] = A;
                iFlag = false;
              }

              else if (oFlag) {
                if (latchFlag) {
                  latchFlag = false;
                  if (A) oState[n] = true;

                }//END if(latchFlag)
                else if (unlatchFlag) {
                  unlatchFlag = false;
                  if (A) oState[n] = false;
                }

                else if (negateFlag) {
                  negateFlag = false;
                  if (A) oState[n] = false;
                  else oState[n] = true;
                }

                else if ((!latchFlag) && (!unlatchFlag) && (!negateFlag)) {
                  oState[n] = A;
                }
                oFlag = false;
              }

              else if (oFlag1) {
                if (latchFlag) {
                  latchFlag = false;
                  if (A) oState[10 + n] = true;

                }//END if(latchFlag)
                else if (unlatchFlag) {
                  unlatchFlag = false;
                  if (A) oState[10 + n] = false;
                }

                else if (negateFlag) {
                  negateFlag = false;
                  if (A) oState[10 + n] = false;
                  else oState[10 + n] = true;
                }

                else if ((!latchFlag) && (!unlatchFlag) && (!negateFlag)) {
                  oState[10 + n] = A;
                }
                oFlag1 = false;
              }

              else if (vFlag) {
                if (latchFlag) {
                  latchFlag = false;
                  if (A) vState[n] = true;

                }//END if(latchFlag)
                else if (unlatchFlag) {
                  unlatchFlag = false;
                  if (A) vState[n] = false;
                }

                else if (negateFlag) {
                  negateFlag = false;
                  if (A) vState[n] = false;
                  else vState[n] = true;
                }

                else if ((!latchFlag) && (!unlatchFlag) && (!negateFlag)) {
                  vState[n] = A;
                }
                vFlag = false;
              }

              else if (vFlag1) {
                if (latchFlag) {
                  latchFlag = false;
                  if (A) vState[10 + n] = true;

                }//END if(latchFlag)
                else if (unlatchFlag) {
                  unlatchFlag = false;
                  if (A) vState[10 + n] = false;
                }

                else if (negateFlag) {
                  negateFlag = false;
                  if (A) vState[10 + n] = false;
                  else vState[10 + n] = true;
                }

                else if ((!latchFlag) && (!unlatchFlag) && (!negateFlag)) {
                  vState[10 + n] = A;
                }
                vFlag1 = false;
              }

              else if (vFlag2) {
                if (latchFlag) {
                  latchFlag = false;
                  if (A) vState[20 + n] = true;

                }//END if(latchFlag)
                else if (unlatchFlag) {
                  unlatchFlag = false;
                  if (A) vState[20 + n] = false;
                }

                else if (negateFlag) {
                  negateFlag = false;
                  if (A) vState[20 + n] = false;
                  else vState[20 + n] = true;
                }

                else if ((!latchFlag) && (!unlatchFlag) && (!negateFlag)) {
                  vState[20 + n] = A;
                }
                vFlag2 = false;
              }

              else if (vFlag3) {
                if (latchFlag) {
                  latchFlag = false;
                  if (A) vState[30 + n] = true;

                }//END if(latchFlag)
                else if (unlatchFlag) {
                  unlatchFlag = false;
                  if (A) vState[30 + n] = false;
                }

                else if (negateFlag) {
                  negateFlag = false;
                  if (A) vState[30 + n] = false;
                  else vState[30 + n] = true;
                }

                else if ((!latchFlag) && (!unlatchFlag) && (!negateFlag)) {
                  vState[30 + n] = A;
                }
                vFlag3 = false;
              }

              else if (vFlag4) {
                if (latchFlag) {
                  latchFlag = false;
                  if (A) vState[40 + n] = true;

                }//END if(latchFlag)
                else if (unlatchFlag) {
                  unlatchFlag = false;
                  if (A) vState[40 + n] = false;
                }

                else if (negateFlag) {
                  negateFlag = false;
                  if (A) vState[40 + n] = false;
                  else vState[40 + n] = true;
                }

                else if ((!latchFlag) && (!unlatchFlag) && (!negateFlag)) {
                  vState[40 + n] = A;
                }
                vFlag4 = false;
              }

              else if (vFlag5) {
                if (latchFlag) {
                  latchFlag = false;
                  if (A) vState[50 + n] = true;

                }//END if(latchFlag)
                else if (unlatchFlag) {
                  unlatchFlag = false;
                  if (A) vState[50 + n] = false;
                }

                else if (negateFlag) {
                  negateFlag = false;
                  if (A) vState[50 + n] = false;
                  else vState[50 + n] = true;
                }

                else if ((!latchFlag) && (!unlatchFlag) && (!negateFlag)) {
                  vState[50 + n] = A;
                }
                vFlag5 = false;
              }

              else if (vFlag6) {
                if (latchFlag) {
                  latchFlag = false;
                  if (A) vState[60 + n] = true;

                }//END if(latchFlag)
                else if (unlatchFlag) {
                  unlatchFlag = false;
                  if (A) vState[60 + n] = false;
                }

                else if (negateFlag) {
                  negateFlag = false;
                  if (A) vState[60 + n] = false;
                  else vState[60 + n] = true;
                }

                else if ((!latchFlag) && (!unlatchFlag) && (!negateFlag)) {
                  vState[60 + n] = A;
                }
                vFlag6 = false;
              }

              //coilFlag=false;
            }//END if(coilFlag)

            else if ((!coilFlag) && (!counterFlag) && (!counterCTUDFlag) && (!timerFlag)) {

              if ((iFlag)) {
                if ((notFlag) && (!risingFlag) && (!fallingFlag)) {
                  if (iState[n]) X = false;
                  else  X = true;
                  //X = ~(iState[n]);
                  A = A & X;
                  notFlag = false;
                }//END IF (notFlag)
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (iUpEdge[n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (iDownEdge[n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (iState[n] & A);
                }
                iFlag = false;
              }//END if(iFlag)

              else if ((iFlag1)) {
                if (notFlag) {
                  if (iState[10 + n]) X = false;
                  else  X = true;
                  //X = ~(iState[10+n]);
                  A = A & X;
                  notFlag = false;
                }//END IF (notFlag)
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (iUpEdge[10 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (iDownEdge[10 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (iState[10 + n] & A);
                }//END else
                iFlag1 = false;
              }//END if(iFlag)

              else if ((oFlag)) {
                n = dataByte;
                if ((notFlag) && (!risingFlag) && (!fallingFlag)) {
                  if (oState[n]) X = false;
                  else  X = true;
                  //X = ~(oState[n]);
                  A = A & X;
                  notFlag = false;
                }//END IF (notFlag)
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (oUpEdge[n]) X = true;
                  else X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (oDownEdge[n]) X = true;
                  else X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (oState[n] & A);
                }//END else
                oFlag = false;
              }//END if(oFlag)

              else if ((oFlag1)) {
                n = dataByte;
                if ((notFlag) && (!risingFlag) && (!fallingFlag)) {
                  if (oState[10 + n]) X = false;
                  else  X = true;
                  //X = ~(oState[10+n]);
                  A = A & X;
                  notFlag = false;
                }//END IF (notFlag)
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (oUpEdge[10 + n]) X = true;
                  else X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (oDownEdge[10 + n]) X = true;
                  else X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (oState[10 + n] & A);
                }//END else
                oFlag1 = false;
              }//END if(oFlag)

              else if ((vFlag)) {
                n = dataByte;
                if ((notFlag) && (!risingFlag) && (!fallingFlag)) {
                  if (vState[n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag = false;
                }//END
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (vUpEdge[n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (vDownEdge[n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (vState[n] & A);
                }//END else
                vFlag = false;
              }//END if(vFlag)

              else if ((vFlag1)) {
                n = dataByte;
                if ((notFlag) && (!risingFlag) && (!fallingFlag)) {
                  if (vState[10 + n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag = false;
                }//END IF (notFlag)
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (vUpEdge[10 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (vDownEdge[10 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (vState[10 + n] & A);
                }//END else
                vFlag1 = false;
              }//END if(vFlag1)

              else if ((vFlag2)) {
                n = dataByte;
                if ((notFlag) && (!risingFlag) && (!fallingFlag)) {
                  if (vState[20 + n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag = false;
                }//END IF (notFlag)
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (vUpEdge[20 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (vDownEdge[20 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (vState[20 + n] & A);
                }//END else
                vFlag2 = false;
              }//END if(vFlag2)

              else if ((vFlag3)) {
                n = dataByte;
                if ((notFlag) && (!risingFlag) && (!fallingFlag)) {
                  if (vState[30 + n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag = false;
                }//END IF (notFlag)
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (vUpEdge[30 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (vDownEdge[30 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (vState[30 + n] & A);
                }//END else
                vFlag3 = false;
              }//END if(vFlag3)

              else if ((vFlag4)) {
                n = dataByte;
                if ((notFlag) && (!risingFlag) && (!fallingFlag)) {
                  if (vState[40 + n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag = false;
                }//END IF (notFlag)
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (vUpEdge[40 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (vDownEdge[40 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (vState[40 + n] & A);
                }//END else
                vFlag4 = false;
              }//END if(vFlag4)

              else if ((vFlag5)) {
                n = dataByte;
                if ((notFlag) && (!risingFlag) && (!fallingFlag)) {
                  if (vState[50 + n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag = false;
                }//END IF (notFlag)
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (vUpEdge[50 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (vDownEdge[50 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (vState[50 + n] & A);
                }//END else
                vFlag5 = false;
              }//END if(vFlag5)

              else if ((vFlag6)) {
                n = dataByte;
                if ((notFlag) && (!risingFlag) && (!fallingFlag)) {
                  if (vState[60 + n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag = false;
                }//END IF (notFlag)
                else if ((!notFlag) && (risingFlag) && (!fallingFlag)) {
                  risingFlag = false;

                  if (vUpEdge[60 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (fallingFlag)) {
                  fallingFlag = false;

                  if (vDownEdge[60 + n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if ((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A = (vState[60 + n] & A);
                }//END else
                vFlag6 = false;
              }//END if(vFlag6)

            }//END else if(~cFlag)


            else if ((!coilFlag) && (counterFlag) && (!counterCTUDFlag) && (!timerFlag)) { //Counter Section CTU

              bool tempNo;
              bool tempReset;
              bool tempDn;

              switch (myCounter[n].no) {
                case 0:
                  tempNo = vState[myCounter[n].noIndex];
                  break;

                case 1:
                  tempNo = iState[myCounter[n].noIndex];
                  break;

                case 2:
                  tempNo = oState[myCounter[n].noIndex];
                  break;
              }

              switch (myCounter[n].reset) {
                case 0:
                  tempReset = vState[myCounter[n].resetIndex];
                  break;

                case 1:
                  tempReset = iState[myCounter[n].resetIndex];
                  break;

                case 2:
                  tempReset = oState[myCounter[n].resetIndex];
                  break;
              }

              if (myCounter[n].dn) tempDn = oState[myCounter[n].dnIndex];
              else tempDn = vState[myCounter[n].dnIndex];


              //PUT the same code of the PLC Ladder

              if ((tempNo) && (!tempReset)) {

                if (!myCounter[n].resetNo) {

                  if ((tempNo) && (!myCounter[n].countNo)) {

                    dintValues[myCounter[n].countIndex]++;

                    myCounter[n].countNo = true;
                    bool returned = setDnCounter(n, tempDn);
                    if (!myCounter[n].dnNull) {
                      if (myCounter[n].dn) oState[myCounter[n].dnIndex] = returned;
                      else vState[myCounter[n].dnIndex] = returned;
                    }

                    if (!myCounter[n].countUpNull) {
                      if (myCounter[n].countUp)  oState[myCounter[n].countUpIndex] = true;
                      else  vState[myCounter[n].countUpIndex] = true;
                    }



                  }
                }//END if(!myCounter[n].resetNo)
              }//END if((tempNo)&&(!tempReset))

              else if ((!tempNo) && (myCounter[n].countNo))  {
                myCounter[n].countNo = false;

                if (!myCounter[n].countUpNull) {
                  if (myCounter[n].countUp)  oState[myCounter[n].countUpIndex] = false;
                  else  vState[myCounter[n].countUpIndex] = false;
                }
              }

              if ((tempReset) && (!myCounter[n].resetNo)) {
                myCounter[n].resetNo = true;
                bool returned = setDnCounter(n, tempDn);
                if (!myCounter[n].dnNull) {
                  if (myCounter[n].dn) oState[myCounter[n].dnIndex] = returned;
                  else vState[myCounter[n].dnIndex] = returned;
                }
                dintValues[myCounter[n].countIndex] = 0;

                if (!myCounter[n].countUpNull) {
                  if (myCounter[n].countUp)  oState[myCounter[n].countUpIndex] = false;
                  else  vState[myCounter[n].countUpIndex] = false;
                }
              }

              else if ((!tempReset) && (myCounter[n].resetNo)) {
                myCounter[n].resetNo = false;

              }


              counterFlag = false;


            }//END Counter Section

            else if ((!coilFlag) && (!counterFlag) && (counterCTUDFlag) && (!timerFlag)) { //Counter Section CTUD


              bool tempNo;
              bool tempReset;
              bool tempNoDown;
              bool tempDn;

              switch (myCounterCTUD[n].no) {
                case 0:
                  tempNo = vState[myCounterCTUD[n].noIndex];
                  break;

                case 1:
                  tempNo = iState[myCounterCTUD[n].noIndex];
                  break;

                case 2:
                  tempNo = oState[myCounterCTUD[n].noIndex];
                  break;
              }

              switch (myCounterCTUD[n].noDown) {
                case 0:
                  tempNoDown = vState[myCounterCTUD[n].noDownIndex];
                  break;

                case 1:
                  tempNoDown = iState[myCounterCTUD[n].noDownIndex];
                  break;

                case 2:
                  tempNoDown = oState[myCounterCTUD[n].noDownIndex];
                  break;
              }

              switch (myCounterCTUD[n].reset) {
                case 0:
                  tempReset = vState[myCounterCTUD[n].resetIndex];
                  break;

                case 1:
                  tempReset = iState[myCounterCTUD[n].resetIndex];
                  break;

                case 2:
                  tempReset = oState[myCounterCTUD[n].resetIndex];
                  break;
              }


              if (myCounterCTUD[n].dn) tempDn = oState[myCounterCTUD[n].dnIndex];
              else tempDn = vState[myCounterCTUD[n].dnIndex];


              //PUT the same code of the PLC Ladder

              if ((tempNo) && (!tempReset)) {
                if (!myCounterCTUD[n].resetNo) {
                  if ((tempNo) && (!myCounterCTUD[n].countNo)) {

                    dintValues[myCounterCTUD[n].countIndex]++;

                    myCounterCTUD[n].countNo = true;
                    bool returned = setDnCounterCTUD(n, tempDn);
                    if (!myCounterCTUD[n].dnNull) {
                      if (myCounterCTUD[n].dn) oState[myCounterCTUD[n].dnIndex] = returned;
                      else vState[myCounterCTUD[n].dnIndex] = returned;
                    }

                    if (!myCounterCTUD[n].countUpNull) {
                      if (myCounterCTUD[n].countUp)  oState[myCounterCTUD[n].countUpIndex] = true;
                      else  vState[myCounterCTUD[n].countUpIndex] = true;
                    }

                  }
                }//END if(!myCounterCTUD[n].resetNo)
              }//END if((tempNo)&&(!tempReset))

              else if ((!tempNo) && (myCounterCTUD[n].countNo))  {
                myCounterCTUD[n].countNo = false;

                if (!myCounterCTUD[n].countUpNull) {
                  if (myCounterCTUD[n].countUp)  oState[myCounterCTUD[n].countUpIndex] = false;
                  else  vState[myCounterCTUD[n].countUpIndex] = false;
                }
              }//END else if((!tempNo) && (myCounterCTUD[n].countNo)


              else if ((tempNoDown) && (!tempReset)) {
                if (!myCounterCTUD[n].resetNo) {
                  if (!myCounterCTUD[n].countNoDown) {

                    dintValues[myCounterCTUD[n].countIndex]--;
                    myCounterCTUD[n].countNoDown = true;
                    bool returned = setDnCounterCTUD(n, tempDn);
                    if (!myCounterCTUD[n].dnNull) {
                      if (myCounterCTUD[n].dn) oState[myCounterCTUD[n].dnIndex] = returned;
                      else vState[myCounterCTUD[n].dnIndex] = returned;
                    }

                    if (!myCounterCTUD[n].countDownNull) {
                      if (myCounterCTUD[n].countDown)  oState[myCounterCTUD[n].countDownIndex] = true;
                      else  vState[myCounterCTUD[n].countDownIndex] = true;
                    }

                  }//END if(!myCounterCTUD[n].countNoDown)
                }//END if(!myCounterCTUD[n].resetNo)
              }//END else if((tempNoDown) && (!tempReset))


              else if ((!tempNoDown) && (myCounterCTUD[n].countNoDown)) {
                myCounterCTUD[n].countNoDown = false;

                if (!myCounterCTUD[n].countDownNull) {
                  if (myCounterCTUD[n].countDown)  oState[myCounterCTUD[n].countDownIndex] = false;
                  else  vState[myCounterCTUD[n].countDownIndex] = false;
                }
              }//END else if((!tempNoDown) && (myCounterCTUD[n].countNoDown))

              if ((tempReset) && (!myCounterCTUD[n].resetNo)) {
                myCounterCTUD[n].resetNo = true;
                bool returned = setDnCounterCTUD(n, tempDn);
                if (!myCounterCTUD[n].dnNull) {
                  if (myCounterCTUD[n].dn) oState[myCounterCTUD[n].dnIndex] = returned;
                  else vState[myCounterCTUD[n].dnIndex] = returned;
                }
                dintValues[myCounterCTUD[n].countIndex] = 0;

                if (!myCounterCTUD[n].countUpNull) {
                  if (myCounterCTUD[n].countUp)  oState[myCounterCTUD[n].countUpIndex] = false;
                  else  vState[myCounterCTUD[n].countUpIndex] = false;
                }
              }

              else if ((!tempReset) && (myCounterCTUD[n].resetNo)) {
                myCounterCTUD[n].resetNo = false;

              }


              counterCTUDFlag = false;


            }//END Counter Section



            else if ((!coilFlag) && (!counterFlag) && (!counterCTUDFlag) && (timerFlag)) { //Timer Section

              bool tempNo;

              unsigned long tempTime;

              switch (myTimer[n].no) {
                case 0:
                  tempNo = vState[myTimer[n].noIndex];
                  break;

                case 1:
                  tempNo = iState[myTimer[n].noIndex];
                  break;

                case 2:
                  tempNo = oState[myTimer[n].noIndex];
                  break;
              }


              switch (myTimer[n].typeTimer) {
                case 1:                                                           //TON
                  if (tempNo) {
                    if (!myTimer[n].activeTimer) {
                      myTimer[n].activeTimer = true;

                      if (!myTimer[n].ttNull) {
                        if (myTimer[n].tt) oState[myTimer[n].ttIndex] = true;
                        else vState[myTimer[n].ttIndex] = true;
                      }

                      myTimer[n].startTime = millis();
                    }
                  }

                  else {
                    myTimer[n].activeTimer = false;


                    if (!myTimer[n].ttNull) {
                      if (myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                      else vState[myTimer[n].ttIndex] = false;
                    }

                    if (!myTimer[n].dnNull) {

                      if (myTimer[n].dn) oState[myTimer[n].dnIndex] = false;
                      else vState[myTimer[n].dnIndex] = false;

                      dintValues[myTimer[n].currentTimeIndex] = 0;

                    }
                  }

                  if (myTimer[n].activeTimer) {
                    tempTime = millis();
                    tempTime = tempTime - myTimer[n].startTime;

                    if (tempTime < 0) {
                      myTimer[n].startTime = millis();
                      tempTime = millis();
                      tempTime = tempTime - myTimer[n].startTime;
                    }

                    dintValues[myTimer[n].currentTimeIndex] = (int32_t)(tempTime * (0.001f));

                    uint32_t tempPreset = 0;
                    if (myTimer[n].isPoint1)
                      tempPreset = (uint32_t)(((0.1) * (dintValues[myTimer[n].presetIndex])) * (1000L));
                    else
                      tempPreset = (uint32_t)((dintValues[myTimer[n].presetIndex]) * (1000L));

                    if (tempTime >= tempPreset) {
                      delay(20);

                      if (!myTimer[n].ttNull) {
                        if (myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                        else vState[myTimer[n].ttIndex] = false;
                      }

                      if (!myTimer[n].dnNull) {
                        if (myTimer[n].dn) oState[myTimer[n].dnIndex] = true;
                        else vState[myTimer[n].dnIndex] = true;
                      }
                    }
                  }
                  break;

                case 2:                                                                 //TOF
                  if (tempNo) {
                    if (!myTimer[n].enableNo) {
                      myTimer[n].enableNo = true;

                      if (!myTimer[n].ttNull) {
                        if (myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                        else vState[myTimer[n].ttIndex] = false;
                      }

                      if (!myTimer[n].dnNull) {
                        if (myTimer[n].dn) oState[myTimer[n].dnIndex] = true;
                        else vState[myTimer[n].dnIndex] = true;
                      }
                    }//END if(!myTimer[n].enableNo)

                    else  {
                      myTimer[n].activeTimer = false;
                      dintValues[myTimer[n].currentTimeIndex] = 0;
                    }
                  }//END if(tempNo

                  else {
                    if (myTimer[n].enableNo) {
                      myTimer[n].enableNo = false;

                      if (!myTimer[n].ttNull) {
                        if (myTimer[n].tt) oState[myTimer[n].ttIndex] = true;
                        else vState[myTimer[n].ttIndex] = true;
                      }

                      myTimer[n].activeTimer = true;
                      myTimer[n].startTime = millis();
                    }//END if(myTimer[n].enableNo)
                  }//END else

                  if (myTimer[n].activeTimer) {
                    tempTime = millis();
                    tempTime = tempTime - myTimer[n].startTime;

                    if (tempTime < 0) {
                      myTimer[n].startTime = millis();
                      tempTime = millis();
                      tempTime = tempTime - myTimer[n].startTime;
                    }

                    dintValues[myTimer[n].currentTimeIndex] = (int32_t)(tempTime * (0.001f));

                    uint32_t tempPreset = 0;
                    if (myTimer[n].isPoint1)
                      tempPreset = (uint32_t)(((0.1) * (dintValues[myTimer[n].presetIndex])) * (1000L));
                    else
                      tempPreset = (uint32_t)((dintValues[myTimer[n].presetIndex]) * (1000L));

                    if (tempTime >= tempPreset) {
                      dintValues[myTimer[n].currentTimeIndex] = 0;
                      delay(20);
                      if (!myTimer[n].ttNull) {
                        if (myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                        else vState[myTimer[n].ttIndex] = false;
                      }

                      if (!myTimer[n].dnNull) {
                        if (myTimer[n].dn) oState[myTimer[n].dnIndex] = false;
                        else vState[myTimer[n].dnIndex] = false;
                      }

                      myTimer[n].activeTimer = false;
                    }//END if(tempTime >= myTimer[n].preset)

                  }//END if(myTimer[n].activeTimer)
                  break;


                case 3:                                                                 //RTO

                  bool tempReset;

                  switch (myTimer[n].reset) {
                    case 0:
                      tempReset = vState[myTimer[n].resetIndex];
                      break;

                    case 1:
                      tempReset = iState[myTimer[n].resetIndex];
                      break;

                    case 2:
                      tempReset = oState[myTimer[n].resetIndex];
                      break;
                  }


                  if (tempNo) {
                    if ((!myTimer[n].activeTimer) && (!myTimer[n].timerDone)) {
                      myTimer[n].activeTimer = true;

                      if (!myTimer[n].ttNull) {
                        if (myTimer[n].tt) oState[myTimer[n].ttIndex] = true;
                        else vState[myTimer[n].ttIndex] = true;
                      }
                      myTimer[n].startTime = millis();
                    }
                  }//END if(tempNo)

                  else {
                    if (myTimer[n].activeTimer) {
                      myTimer[n].activeTimer = false;

                      if (!myTimer[n].ttNull) {
                        if (myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                        else vState[myTimer[n].ttIndex] = false;
                      }
                      long tempTime = millis();
                      int32_t tempValue = dintValues[myTimer[n].currentTimeIndex] * (1000L);
                      tempValue = tempValue + tempTime - myTimer[n].startTime;

                      int32_t tempStartCheck = tempTime - myTimer[n].startTime;
                      if (tempStartCheck < 0) {
                        myTimer[n].startTime = millis();
                        tempTime = millis();
                        tempValue = tempValue + tempTime - myTimer[n].startTime;
                      }

                      dintValues[myTimer[n].currentTimeIndex] = (int32_t)tempValue * (0.001f);

                    }//END if(mytimer[n].activeTimer)


                  }//END else

                  if (tempReset) {
                    myTimer[n].activeTimer = false;
                    myTimer[n].enableNo = false;

                    if (!myTimer[n].ttNull) {
                      if (myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                      else vState[myTimer[n].ttIndex] = false;
                    }

                    if (!myTimer[n].dnNull) {
                      if (myTimer[n].dn) oState[myTimer[n].dnIndex] = false;
                      else vState[myTimer[n].dnIndex] = false;
                    }

                    dintValues[myTimer[n].currentTimeIndex] = 0;
                    myTimer[n].timerDone = false;
                  }


                  if (myTimer[n].activeTimer) {
                    tempTime = millis();
                    tempTime = tempTime - myTimer[n].startTime;
                    int32_t tempValue = dintValues[myTimer[n].currentTimeIndex] * (1000L);
                    tempTime += tempValue;

                    uint32_t tempPreset = 0;
                    if (myTimer[n].isPoint1)
                      tempPreset = (uint32_t)(((0.1) * (dintValues[myTimer[n].presetIndex])) * (1000L));
                    else
                      tempPreset = (uint32_t)((dintValues[myTimer[n].presetIndex]) * (1000L));

                    if (tempTime >= tempPreset) {

                      delay(20);
                      if (!myTimer[n].ttNull) {
                        if (myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                        else vState[myTimer[n].ttIndex] = false;
                      }//END if(!myTimer[n].ttNull)

                      if (!myTimer[n].dnNull) {
                        if (myTimer[n].dn) oState[myTimer[n].dnIndex] = true;
                        else vState[myTimer[n].dnIndex] = true;
                      }

                      myTimer[n].timerDone = true;

                    }//END if(tempTime >= myTimer[n].preset)
                  }//END if(myTimer[n].activeTimer)

                  break;


              }//END switch(myTimer[n].typeTimer)

              timerFlag = false;
            }//END Timer Section

          }//END if(dataByte<15)
      }//END Switch
    }//END ELSE !INFUNCTION

    PC++;
    dataByte = EEPROM.read(PC);

  }//END while(dataByte!=254)

  //END OF Z
  //Serial.println(" ");

}//END interpreter()

void updateRegisters() {

  for (int i = 0; i < 3; i++) { 
    if (iMask[i]) {         //input

      if ((!iPreState[i]) && (iState[i]))
        iUpEdge[i] = true;
      else if ((iPreState[i]) && (!iState[i]))
        iDownEdge[i] = true;
      else {
        iUpEdge[i] = false;
        iDownEdge[i] = false;
      }

      iPreState[i] = iState[i];
      iState[i] = digitalRead(iPins[i]);
      //delay(10);
    }

  }//END FOR input

  for (int i = 0; i < 4; i++) {
    if (oMask[i]) {         //output

      if ((!oPreState[i]) && (oState[i]))
        oUpEdge[i] = true;
      else if ((oPreState[i]) && (!oState[i]))
        oDownEdge[i] = true;
      else {
        oUpEdge[i] = false;
        oDownEdge[i] = false;
      }

      oPreState[i] = oState[i];
      if (analogOutputs[i] == 0)
        digitalWrite(oPins[i], oState[i]);

      //delay(10);
    }
  }

  for (int i = 0; i < 64; i++) {
    if (vMask[i]) {

      if ((!vPreState[i]) && (vState[i]))
        vUpEdge[i] = true;
      else if ((vPreState[i]) && (!vState[i]))
        vDownEdge[i] = true;
      else {
        vUpEdge[i] = false;
        vDownEdge[i] = false;
      }

      vPreState[i] = vState[i];
    }
  }
}//END updateRegisters()

void cleanOutputs() {

  for (int i = 0; i < 3; i++) { //Input   
    preIState[i] = true;
    
    iState[i] = false;

    iPreState[i] = false;

    iUpEdge[i] = false;

    iDownEdge[i] = false;

    analogIEnabled[i] = false;

    analogInputs[i] = 0;

    preAnalogInputs[i] = 1;
  }

  for (int i = 0; i < 4; i++) { //Output
    preOState[i] = true;

    digitalWrite(oPins[i], false);

    oState[i] = false;

    oPreState[i] = false;

    oUpEdge[i] = false;

    oDownEdge[i] = false;

    analogOutputs[i] = 0;
    
    preAnalogOutputs[i] = 1;

    pwmOutput[i] = false;

    servoOutput[i] = false;
  }

  for (int i = 0; i < 64; i++) {
    vState[i] = false;

    vPreState[i] = false;

    vUpEdge[i] = false;

    vDownEdge[i] = false;
  }
}

//fOR THE EEPROM 255 = true, 0 = false
void initProgram() {
  if (EEPROM.read(3959) == '2') { //memory status flag

    cleanOutputs();

    inputSize = 0;
    outputSize = 0;
    for (int i = 0; i < 3; i++) { // input variables mask 
      int tempVar = EEPROM.read(3960 + i); //input
      if (tempVar == 126) {
        iMask[i] = true;
        inputSize++;
      }

    }//END FOR i mask

    for (int i = 0; i < 4; i++) { // output variables mask
      int tempVar = EEPROM.read(3976 + i); //output
      if (tempVar == 126) {
        oMask[i] = true;
        outputSize++;
      }
    }//END FOR o mask

    virtualSize = 0;
    for (int i = 0; i < 64; i++) { //virtual variables mask
      int tempVar = EEPROM.read(3992 + i);
      if (tempVar == 126) {
        vMask[i] = true;
        virtualSize++;
      }
    }//END FOR virtual mask

    //INIT INT VARIABLES
    int tempPC = 2866;
    int32_t tempA = 0;
    int32_t tempB = 0;
    int intIndex = 0;

    while (tempPC < 2930) {
      tempA = (int32_t)EEPROM.read(tempPC);


      tempA = (int32_t)tempA << 8;
      tempPC++;

      tempA = tempA | (int32_t)EEPROM.read(tempPC);
      tempPC++;

      intValues[intIndex] = (int16_t)tempA;

      intIndex++;

    }

    //INIT DINT VARIABLES
    tempPC = 2931;
    tempA = 0;
    tempB = 0;
    intIndex = 0;

    while (tempPC < 3155) {
      tempA = EEPROM.read(tempPC);
      tempA = ((int32_t)tempA) << 24;
      tempPC++;

      tempB = EEPROM.read(tempPC);
      tempB = ((int32_t)tempB) << 16;
      tempA = tempA | tempB;
      tempPC++;

      tempB = EEPROM.read(tempPC);
      tempB = ((int32_t)tempB) << 8;
      tempA = tempA | tempB;
      tempPC++;

      tempA = tempA | ((int32_t)EEPROM.read(tempPC));
      tempPC++;

      dintValues[intIndex] = tempA;
      intIndex++;

    }

    //INIT REAL VARIABLES
    tempPC = 3156;
    tempA = 0;
    tempB = 0;
    intIndex = 0;

    floatValue tempFloat;
    while (tempPC < 3284) {
      tempFloat.bytes[3] = EEPROM.read(tempPC);
      tempPC++;

      tempFloat.bytes[2] = EEPROM.read(tempPC);
      tempPC++;

      tempFloat.bytes[1] = EEPROM.read(tempPC);
      tempPC++;

      tempFloat.bytes[0] = EEPROM.read(tempPC);
      tempPC++;

      realValues[intIndex] = tempFloat.val;

      intIndex++;

    }



    for (int i = 0; i < 16; i++) { //Loading Counters 16 COUNTERS
      int tempMulti = 3574 + (16 * i);
      char tempCounterValue = EEPROM.read(tempMulti);

      if (tempCounterValue == 'B') { //IF THE COUNTER IS ACTIVE  //CTU

        int tempFlag = EEPROM.read(tempMulti + 1);
        if (tempFlag == 'I') {
          myCounter[i].no = 1;
          myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
        }
        else if (tempFlag == 'i') {
          myCounter[i].no = 1;
          myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
        }
        else if (tempFlag == 'O') {
          myCounter[i].no = 2;
          myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
        }
        else if (tempFlag == 'k') {
          myCounter[i].no = 2;
          myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
        }
        else {
          myCounter[i].no = 0;
          switch (tempFlag) {
            case 'V':
              myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
              break;

            case 'v':
              myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
              break;

            case 'U':
              myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 20;
              break;

            case 'u':
              myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 30;
              break;

            case 'A':
              myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 40;
              break;

            case 'a':
              myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 50;
              break;

            case 'c':
              myCounter[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 60;
              break;
          }//END switch
        }


        tempFlag = EEPROM.read(tempMulti + 3);
        if (tempFlag == 'I') {
          myCounter[i].reset = 1;
          myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48;
        }
        else if (tempFlag == 'i') {
          myCounter[i].reset = 1;
          myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
        }
        else if (tempFlag == 'O') {
          myCounter[i].reset = 2;
          myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48;
        }
        else if (tempFlag == 'k') {
          myCounter[i].reset = 2;
          myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
        }
        else {
          myCounter[i].reset = 0;
          switch (tempFlag) {
            case 'V':
              myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48;
              break;

            case 'v':
              myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
              break;

            case 'U':
              myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 20;
              break;

            case 'u':
              myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 30;
              break;

            case 'A':
              myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 40;
              break;

            case 'a':
              myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 50;
              break;

            case 'c':
              myCounter[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 60;
              break;
          }//END switch
        }


        tempFlag = EEPROM.read(tempMulti + 5);
        if (tempFlag == 's') { //null
          myCounter[i].dnNull = true;
        }
        else {
          myCounter[i].dnNull = false;
          if (tempFlag == 'O') {
            myCounter[i].dn = true;
            myCounter[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48;
          }
          else if (tempFlag == 'k') {
            myCounter[i].dn = true;
            myCounter[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 10;
          }
          else {
            myCounter[i].dn = false;
            switch (tempFlag) {
              case 'V':
                myCounter[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48;
                break;

              case 'v':
                myCounter[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 10;
                break;

              case 'U':
                myCounter[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 20;
                break;

              case 'u':
                myCounter[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 30;
                break;

              case 'A':
                myCounter[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 40;
                break;

              case 'a':
                myCounter[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 50;
                break;

              case 'c':
                myCounter[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 60;
                break;

            }//END switch
          }
        }//END else

        tempFlag = EEPROM.read(tempMulti + 7);
        if (tempFlag == 's') {
          myCounter[i].countUpNull = true;
        }
        else {
          myCounter[i].countUpNull = false;
          if (tempFlag == 'O') {
            myCounter[i].countUp = true;
            myCounter[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48;
          }
          else if (tempFlag == 'k') {
            myCounter[i].countUp = true;
            myCounter[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 10;
          }
          else {
            myCounter[i].countUp = false;
            switch (tempFlag) {
              case 'V':
                myCounter[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48;
                break;

              case 'v':
                myCounter[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 10;
                break;

              case 'U':
                myCounter[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 20;
                break;

              case 'u':
                myCounter[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 30;
                break;

              case 'A':
                myCounter[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 40;
                break;

              case 'a':
                myCounter[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 50;
                break;

              case 'c':
                myCounter[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 60;
                break;
            }//END switch
          }
        }//END else


        tempFlag = EEPROM.read(tempMulti + 9);
        if (tempFlag == 'P') {
          myCounter[i].presetIndex = EEPROM.read(tempMulti + 10);
          myCounter[i].countIndex = EEPROM.read(tempMulti + 11);
          myCounter[i].countNo = false;
          myCounter[i].resetNo = false;
        }


      }//END if(244)


      else if (tempCounterValue == 'b') {                                             //CTUD

        int tempFlag = EEPROM.read(tempMulti + 1);
        if (tempFlag == 'I') {
          myCounterCTUD[i].no = 1;
          myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
        }
        else if (tempFlag == 'i') {
          myCounterCTUD[i].no = 1;
          myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
        }
        else if (tempFlag == 'O') {
          myCounterCTUD[i].no = 2;
          myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
        }
        else if (tempFlag == 'k') {
          myCounterCTUD[i].no = 2;
          myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
        }
        else {
          myCounterCTUD[i].no = 0;
          switch (tempFlag) {
            case 'V':
              myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
              break;

            case 'v':
              myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
              break;

            case 'U':
              myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 20;
              break;

            case 'u':
              myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 30;
              break;

            case 'A':
              myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 40;
              break;

            case 'a':
              myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 50;
              break;

            case 'c':
              myCounterCTUD[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 60;
              break;
          }//END switch
        }


        tempFlag = EEPROM.read(tempMulti + 3);
        if (tempFlag == 'I') {
          myCounterCTUD[i].reset = 1;
          myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48;
        }
        else if (tempFlag == 'i') {
          myCounterCTUD[i].reset = 1;
          myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
        }
        else if (tempFlag == 'O') {
          myCounterCTUD[i].reset = 2;
          myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48;
        }
        else if (tempFlag == 'k') {
          myCounterCTUD[i].reset = 2;
          myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
        }
        else {
          myCounterCTUD[i].reset = 0;
          switch (tempFlag) {
            case 'V':
              myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48;
              break;

            case 'v':
              myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
              break;

            case 'U':
              myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 20;
              break;

            case 'u':
              myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 30;
              break;

            case 'A':
              myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 40;
              break;

            case 'a':
              myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 50;
              break;

            case 'c':
              myCounterCTUD[i].resetIndex = (EEPROM.read(tempMulti + 4)) - 48 + 60;
              break;
          }//END switch
        }


        tempFlag = EEPROM.read(tempMulti + 5);
        if (tempFlag == 's') {
          myCounterCTUD[i].dnNull = true;
        }
        else {
          myCounterCTUD[i].dnNull = false;
          if (tempFlag == 'O') {
            myCounterCTUD[i].dn = true;
            myCounterCTUD[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48;
          }
          else if (tempFlag == 'k') {
            myCounterCTUD[i].dn = true;
            myCounterCTUD[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 10;
          }
          else {
            myCounterCTUD[i].dn = false;
            switch (tempFlag) {
              case 'V':
                myCounterCTUD[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48;
                break;

              case 'v':
                myCounterCTUD[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 10;
                break;

              case 'U':
                myCounterCTUD[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 20;
                break;

              case 'u':
                myCounterCTUD[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 30;
                break;

              case 'A':
                myCounterCTUD[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 40;
                break;

              case 'a':
                myCounterCTUD[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 50;
                break;

              case 'c':
                myCounterCTUD[i].dnIndex = (EEPROM.read(tempMulti + 6)) - 48 + 60;
                break;
            }//END switch
          }
        }//END else


        tempFlag = EEPROM.read(tempMulti + 7);
        if (tempFlag == 's') {
          myCounterCTUD[i].countUpNull = true;
        }
        else {
          myCounterCTUD[i].countUpNull = false;
          if (tempFlag == 'O') {
            myCounterCTUD[i].countUp = true;
            myCounterCTUD[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48;
          }
          else if (tempFlag == 'k') {
            myCounterCTUD[i].countUp = true;
            myCounterCTUD[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 10;
          }
          else {
            myCounterCTUD[i].countUp = false;
            switch (tempFlag) {
              case 'V':
                myCounterCTUD[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48;
                break;

              case 'v':
                myCounterCTUD[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 10;
                break;

              case 'U':
                myCounterCTUD[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 20;
                break;

              case 'u':
                myCounterCTUD[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 30;
                break;

              case 'A':
                myCounterCTUD[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 40;
                break;

              case 'a':
                myCounterCTUD[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 50;
                break;

              case 'c':
                myCounterCTUD[i].countUpIndex = (EEPROM.read(tempMulti + 8)) - 48 + 60;
                break;
            }//END switch
          }
        }//END else


        tempFlag = EEPROM.read(tempMulti + 9);
        if (tempFlag == 'P') {

          myCounterCTUD[i].presetIndex = EEPROM.read(tempMulti + 10);
          myCounterCTUD[i].countIndex = EEPROM.read(tempMulti + 11);
          myCounterCTUD[i].countNo = false;
          myCounterCTUD[i].resetNo = false;
        }


        tempFlag = EEPROM.read(tempMulti + 12);                                       //N.O. DOWN
        if (tempFlag == 'I') {
          myCounterCTUD[i].noDown = 1;
          myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48;
        }
        else if (tempFlag == 'i') {
          myCounterCTUD[i].noDown = 1;
          myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48 + 10;
        }
        else if (tempFlag == 'O') {
          myCounterCTUD[i].noDown = 2;
          myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48;
        }
        else if (tempFlag == 'k') {
          myCounterCTUD[i].noDown = 2;
          myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48 + 10;
        }
        else {
          myCounterCTUD[i].noDown = 0;
          switch (tempFlag) {
            case 'V':
              myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48;
              break;

            case 'v':
              myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48 + 10;
              break;

            case 'U':
              myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48 + 20;
              break;

            case 'u':
              myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48 + 30;
              break;

            case 'A':
              myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48 + 40;
              break;

            case 'a':
              myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48 + 50;
              break;

            case 'c':
              myCounterCTUD[i].noDownIndex = (EEPROM.read(tempMulti + 13)) - 48 + 60;
              break;
          }//END switch
        }


        tempFlag = EEPROM.read(tempMulti + 14);
        if (tempFlag == 's') {
          myCounterCTUD[i].countDownNull = true;
        }
        else {
          myCounterCTUD[i].countDownNull = false;
          if (tempFlag == 'O') {
            myCounterCTUD[i].countDown = true;
            myCounterCTUD[i].countDownIndex = (EEPROM.read(tempMulti + 15)) - 48;
          }
          else if (tempFlag == 'k') {
            myCounterCTUD[i].countDown = true;
            myCounterCTUD[i].countDownIndex = (EEPROM.read(tempMulti + 15)) - 48 + 10;
          }
          else {
            myCounterCTUD[i].countDown = false;
            switch (tempFlag) {
              case 'V':
                myCounterCTUD[i].countDownIndex = (EEPROM.read(tempMulti + 15)) - 48;
                break;

              case 'v':
                myCounterCTUD[i].countDownIndex = (EEPROM.read(tempMulti + 15)) - 48 + 10;
                break;

              case 'U':
                myCounterCTUD[i].countDownIndex = (EEPROM.read(tempMulti + 15)) - 48 + 20;
                break;

              case 'u':
                myCounterCTUD[i].countDownIndex = (EEPROM.read(tempMulti + 15)) - 48 + 30;
                break;

              case 'A':
                myCounterCTUD[i].countDownIndex = (EEPROM.read(tempMulti + 15)) - 48 + 40;
                break;

              case 'a':
                myCounterCTUD[i].countDownIndex = (EEPROM.read(tempMulti + 15)) - 48 + 50;
                break;

              case 'c':
                myCounterCTUD[i].countDownIndex = (EEPROM.read(tempMulti + 15)) - 48 + 60;
                break;
            }//END switch
          }
        }//END else

      }//END else if(tempCounterValue == 'b')



    }//END for Counters


    for (int i = 0; i < 16; i++) {                                                     //Loading Timers
      int tempMulti = 3285 + (12 * i);
      char tempCounterValue = EEPROM.read(tempMulti);


      if (tempCounterValue == 't') {                                                                         //TON
        myTimer[i].typeTimer = 1;//TON

        int tempFlag = EEPROM.read(tempMulti + 1);
        if (tempFlag == 'I') {
          myTimer[i].no = 1;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
        }
        else if (tempFlag == 'i') {
          myTimer[i].no = 1;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
        }
        else if (tempFlag == 'O') {
          myTimer[i].no = 2;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
        }
        else if (tempFlag == 'k') {
          myTimer[i].no = 2;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
        }
        else {
          myTimer[i].no = 0;
          switch (tempFlag) {
            case 'V':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
              break;

            case 'v':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
              break;

            case 'U':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 20;
              break;

            case 'u':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 30;
              break;

            case 'A':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 40;
              break;

            case 'a':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 50;
              break;

            case 'c':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 60;
              break;
          }//END switch
        }//END else


        tempFlag = EEPROM.read(tempMulti + 3);
        if (tempFlag == 's') {
          myTimer[i].dnNull = true;
        }
        else {
          myTimer[i].dnNull = false;
          if (tempFlag == 'O') {
            myTimer[i].dn = true;
            myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48;
          }
          else if (tempFlag == 'k') {
            myTimer[i].dn = true;
            myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
          }
          else {
            myTimer[i].dn = false;
            switch (tempFlag) {
              case 'V':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48;
                break;

              case 'v':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
                break;

              case 'U':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 20;
                break;

              case 'u':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 30;
                break;

              case 'A':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 40;
                break;

              case 'a':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 50;
                break;

              case 'c':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 60;
                break;
            }//END switch
          }//END else
        }//END else


        tempFlag = EEPROM.read(tempMulti + 5);
        if (tempFlag == 's') {
          myTimer[i].ttNull = true;
        }
        else {
          myTimer[i].ttNull = false;

          if (tempFlag == 'O') {
            myTimer[i].tt = true;
            myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48;
          }
          else if (tempFlag == 'k') {
            myTimer[i].tt = true;
            myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 10;
          }
          else {
            myTimer[i].tt = false;
            switch (tempFlag) {
              case 'V':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48;
                break;

              case 'v':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 10;
                break;

              case 'U':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 20;
                break;

              case 'u':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 30;
                break;

              case 'A':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 40;
                break;

              case 'a':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 50;
                break;

              case 'c':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 60;
                break;
            }//END switch
          }//END else
        }//END else

        myTimer[i].presetIndex = EEPROM.read(tempMulti + 8);
        myTimer[i].currentTimeIndex = EEPROM.read(tempMulti + 9);


        if (EEPROM.read(tempMulti + 7) == '0')
          myTimer[i].isPoint1 = false;

        else if (EEPROM.read(tempMulti + 7) == '1')
          myTimer[i].isPoint1 = true;


        myTimer[i].enableNo = false;
        myTimer[i].activeTimer = false;
      }



      else if (tempCounterValue == 'T') {                                                                    //TOF
        myTimer[i].typeTimer = 2;//TOF

        int tempFlag = EEPROM.read(tempMulti + 1);
        if (tempFlag == 'I') {
          myTimer[i].no = 1;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
        }
        else if (tempFlag == 'i') {
          myTimer[i].no = 1;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
        }
        else if (tempFlag == 'O') {
          myTimer[i].no = 2;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
        }
        else if (tempFlag == 'k') {
          myTimer[i].no = 2;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
        }
        else {
          myTimer[i].no = 0;
          switch (tempFlag) {
            case 'V':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
              break;

            case 'v':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
              break;

            case 'U':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 20;
              break;

            case 'u':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 30;
              break;

            case 'A':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 40;
              break;

            case 'a':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 50;
              break;

            case 'c':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 60;
              break;
          }//END switch
        }//END else


        tempFlag = EEPROM.read(tempMulti + 3);
        if (tempFlag == 's') {
          myTimer[i].dnNull = true;
        }
        else {
          myTimer[i].dnNull = false;
          if (tempFlag == 'O') {
            myTimer[i].dn = true;
            myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48;
          }
          else if (tempFlag == 'k') {
            myTimer[i].dn = true;
            myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
          }
          else {
            myTimer[i].dn = false;
            switch (tempFlag) {
              case 'V':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48;
                break;

              case 'v':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
                break;

              case 'U':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 20;
                break;

              case 'u':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 30;
                break;

              case 'A':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 40;
                break;

              case 'a':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 50;
                break;

              case 'c':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 60;
                break;
            }//END switch
          }//END else
        }//END else


        tempFlag = EEPROM.read(tempMulti + 5);
        if (tempFlag == 's') {
          myTimer[i].ttNull = true;
        }
        else {
          myTimer[i].ttNull = false;

          if (tempFlag == 'O') {
            myTimer[i].tt = true;
            myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48;
          }
          else if (tempFlag == 'k') {
            myTimer[i].tt = true;
            myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 10;
          }
          else {
            myTimer[i].tt = false;
            switch (tempFlag) {
              case 'V':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48;
                break;

              case 'v':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 10;
                break;

              case 'U':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 20;
                break;

              case 'u':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 30;
                break;

              case 'A':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 40;
                break;

              case 'a':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 50;
                break;

              case 'c':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 60;
                break;
            }//END switch
          }//END else
        }//END else


        myTimer[i].presetIndex = EEPROM.read(tempMulti + 8);
        myTimer[i].currentTimeIndex = EEPROM.read(tempMulti + 9);

        if (EEPROM.read(tempMulti + 7) == '0')
          myTimer[i].isPoint1 = false;

        else if (EEPROM.read(tempMulti + 7) == '1')
          myTimer[i].isPoint1 = true;

        myTimer[i].enableNo = false;
        myTimer[i].activeTimer = false;
      }


      else if (tempCounterValue == 'S') {                                                                    //RTO

        myTimer[i].typeTimer = 3;//RTO

        int tempFlag = EEPROM.read(tempMulti + 1);
        if (tempFlag == 'I') {
          myTimer[i].no = 1;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
        }
        else if (tempFlag == 'i') {
          myTimer[i].no = 1;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
        }
        else if (tempFlag == 'O') {
          myTimer[i].no = 2;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
        }
        else if (tempFlag == 'k') {
          myTimer[i].no = 2;
          myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
        }
        else {
          myTimer[i].no = 0;
          switch (tempFlag) {
            case 'V':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48;
              break;

            case 'v':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 10;
              break;

            case 'U':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 20;
              break;

            case 'u':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 30;
              break;

            case 'A':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 40;
              break;

            case 'a':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 50;
              break;

            case 'c':
              myTimer[i].noIndex = (EEPROM.read(tempMulti + 2)) - 48 + 60;
              break;
          }//END switch
        }//END else


        tempFlag = EEPROM.read(tempMulti + 3);
        if (tempFlag == 's') {
          myTimer[i].dnNull = true;
        }
        else {
          myTimer[i].dnNull = false;
          if (tempFlag == 'O') {
            myTimer[i].dn = true;
            myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48;
          }
          else if (tempFlag == 'k') {
            myTimer[i].dn = true;
            myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
          }
          else {
            myTimer[i].dn = false;
            switch (tempFlag) {
              case 'V':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48;
                break;

              case 'v':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 10;
                break;

              case 'U':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 20;
                break;

              case 'u':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 30;
                break;

              case 'A':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 40;
                break;

              case 'a':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 50;
                break;

              case 'c':
                myTimer[i].dnIndex = (EEPROM.read(tempMulti + 4)) - 48 + 60;
                break;
            }//END switch
          }//END else
        }//END else


        tempFlag = EEPROM.read(tempMulti + 5);
        if (tempFlag == 's') {
          myTimer[i].ttNull = true;
        }
        else {
          myTimer[i].ttNull = false;

          if (tempFlag == 'O') {
            myTimer[i].tt = true;
            myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48;
          }
          else if (tempFlag == 'k') {
            myTimer[i].tt = true;
            myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 10;
          }
          else {
            myTimer[i].tt = false;
            switch (tempFlag) {
              case 'V':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48;
                break;

              case 'v':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 10;
                break;

              case 'U':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 20;
                break;

              case 'u':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 30;
                break;

              case 'A':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 40;
                break;

              case 'a':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 50;
                break;

              case 'c':
                myTimer[i].ttIndex = (EEPROM.read(tempMulti + 6)) - 48 + 60;
                break;
            }//END switch
          }//END else
        }//END else


        myTimer[i].presetIndex = EEPROM.read(tempMulti + 8);
        myTimer[i].currentTimeIndex = EEPROM.read(tempMulti + 9);

        if (EEPROM.read(tempMulti + 7) == '0')
          myTimer[i].isPoint1 = false;

        else if (EEPROM.read(tempMulti + 7) == '1')
          myTimer[i].isPoint1 = true;


        tempFlag = EEPROM.read(tempMulti + 10);
        if (tempFlag == 'I') {
          myTimer[i].reset = 1;
          myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48;
        }
        else if (tempFlag == 'i') {
          myTimer[i].reset = 1;
          myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48 + 10;
        }
        else if (tempFlag == 'O') {
          myTimer[i].reset = 2;
          myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48;
        }
        else if (tempFlag == 'k') {
          myTimer[i].reset = 2;
          myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48 + 10;
        }
        else {
          myTimer[i].reset = 0;
          switch (tempFlag) {
            case 'V':
              myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48;
              break;

            case 'v':
              myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48 + 10;
              break;

            case 'U':
              myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48 + 20;
              break;

            case 'u':
              myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48 + 30;
              break;

            case 'A':
              myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48 + 40;
              break;

            case 'a':
              myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48 + 50;
              break;

            case 'c':
              myTimer[i].resetIndex = (EEPROM.read(tempMulti + 11)) - 48 + 60;
              break;
          }//END switch
        }//END else

        myTimer[i].enableNo = false;
        myTimer[i].activeTimer = false;
        dintValues[myTimer[i].currentTimeIndex] = 0;

      }

    }//END FOR Timers

  }//END IF


  PC = 0; //HERE
  n = 0;
  X = false;
  Y = false;
  notFlag = false;
  fallingFlag = false;
  risingFlag = false;
  iFlag = false;
  iFlag1 = false;
  oFlag = false;
  oFlag1 = false;
  vFlag = false;
  vFlag1 = false;
  vFlag2 = false;
  vFlag3 = false;
  vFlag4 = false;
  vFlag5 = false;
  vFlag6 = false;
  latchFlag = false;
  unlatchFlag = false;
  negateFlag = false;
  coilFlag = false;
  timerFlag = false;
  counterFlag = false;
  counterCTUDFlag = false;

  inFunction = false;


}



void reprogram() {

  if (!clientConnected)
    clientTemp = server.available();


  if (clientTemp) {
    if (clientTemp.connected()) {
      clientTemp.setNoDelay(false);
      clientConnected = true;
    }
  }


  if (clientTemp.available() > 0) {

    if (inReprogramTime) {

      int buffSize = 0;
      buffSize = readMyByteUntil(42, clientTemp);

      if (bufferSerial[0] == 'c') {
        clientTemp.write((uint8_t)'T');
        clientTemp.write((uint8_t)10);
        delay(10);

        countPC = 0;
        countCounter = 0;
        countCounterCTU = 0;
        countCounterCTUD = 0;
        countTimer = 0;
        counterInt = 0;
        counterDint = 0;
        counterReal = 0;

        programming = true;


      } //END if(bufferSerial[0] == 'c')



      if (bufferSerial[0] == 'I') {
        int cursor = 1;
        int newCursor = 0;

        int tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
        cursor += 2;

        if (tempValue > 0) {
          for (int i = 0; i < tempValue; i++ ) {
            //EEPROM.write(REG_IN+i, bufferSerial[cursor+i]);
            preEEPROM[REG_IN + i] = (uint8_t)bufferSerial[cursor + i];
            //delay(50);

            newCursor = cursor + i + 1;
          }

          cursor = newCursor;
        }

       

        tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
        int tempType = bufferSerial[cursor + 2];
        cursor += 3;


        if (tempType == 'O') {
          if (tempValue > 0) {
            for (int i = 0; i < tempValue; i++ ) {
              //EEPROM.write(REG_OUT+i, bufferSerial[cursor+i]);
              preEEPROM[REG_OUT + i] = (uint8_t)bufferSerial[cursor + i];
              //delay(50);
              newCursor = cursor + i + 1;
            }

            cursor = newCursor;
          }
        }


        tempType = bufferSerial[cursor];
        if (tempType == 'X') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'h');
          clientTemp.write((uint8_t)10);

        }

      }//END 'I'

      else if (bufferSerial[0] == 'V') {
        int cursor = 1;
        int newCursor = 0;

        int tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
        cursor += 2;

        if (tempValue > 0) {
          for (int i = 0; i < tempValue; i++ ) {
            //EEPROM.write(REG_VIRTUAL+i, bufferSerial[cursor+i]);
            preEEPROM[REG_VIRTUAL + i] = (uint8_t)bufferSerial[cursor + i];
            //delay(50);
            newCursor = cursor + i + 1;
          }

          cursor = newCursor;
        }


        int tempType = bufferSerial[cursor];
        if (tempType == 'x') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);

        }


      }//END 'V'

      else if (bufferSerial[0] == 'w') {
        
        int cursor = 1;
        int newCursor = 0;
        bool lock = true;
        bool xFound = false;
        int i = 0;

        while (lock) {
          //EEPROM.write(REG_TYPE_SIZE+i, bufferSerial[i+cursor]);
          preEEPROM[REG_TYPE_SIZE + i] = (uint8_t)bufferSerial[i + cursor];
          //delay(120);
          i++;
          newCursor = cursor + i;

          if (i >= buffSize)
            lock = false;

          if (bufferSerial[newCursor] == 'x') {
            xFound = true;
            lock = false;
          }
        }

        if (xFound) {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);
        }



      }//END 'w'


      else if (bufferSerial[0] == 'N') {
        
        int cursor = 1;
        int varSalto = 0;
        int32_t tempA = 0;
        int32_t tempB = 0;

        varSalto = (REG_INT + (2 * counterInt));

        tempB = bufferSerial[cursor + 2] << 14;
        tempA = bufferSerial[cursor + 1] << 7;
        tempB = tempB | tempA | bufferSerial[cursor];

        tempA = tempB & (0xFF);
        //EEPROM.write(varSalto + 1, tempA);
        preEEPROM[varSalto + 1] = (uint8_t)tempA;
        //delay(120);

        tempA = tempB & (0xFF00);
        tempA = tempA >> 8;

        //EEPROM.write(varSalto, tempA);
        preEEPROM[varSalto] = (uint8_t)tempA;
        //delay(120);

        counterInt++;

        int tempType = bufferSerial[cursor + 3];
        if (tempType == 'x') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);

        }


      }//END 'N'

      else if (bufferSerial[0] == 'L') {
        
        int cursor = 1;
        int varSalto = 0;
        int32_t tempA = 0;
        int32_t tempB = 0;

        varSalto = (REG_LONG + (4 * counterDint));

        tempB = (((int32_t)bufferSerial[cursor + 4]) << 28);
        tempA = (((int32_t)bufferSerial[cursor + 3]) << 21);
        tempB = tempB | tempA;
        tempA = (((int32_t)bufferSerial[cursor + 2]) << 14);
        tempB = tempB | tempA;
        tempA = (((int32_t)bufferSerial[cursor + 1]) << 7);
        tempB = tempB | tempA | ((int32_t)bufferSerial[cursor]);

        tempA = tempB & (0xFF);
        //EEPROM.write(varSalto + 3, tempA);
        preEEPROM[varSalto + 3] = (uint8_t)tempA;
        //delay(120);

        tempA = tempB & (0xFF00);
        tempA = tempA >> 8;
        //EEPROM.write(varSalto + 2, tempA);
        preEEPROM[varSalto + 2] = (uint8_t)tempA;
        //delay(120);

        tempA = tempB & (0xFF0000);
        tempA = tempA >> 16;
        //EEPROM.write(varSalto + 1, tempA);
        preEEPROM[varSalto + 1] = (uint8_t)tempA;
        //delay(120);

        tempA = tempB & (0xFF000000);
        tempA = tempA >> 24;
        //EEPROM.write(varSalto, tempA);
        preEEPROM[varSalto] = (uint8_t)tempA;
        //delay(120);

        counterDint++;

        int tempType = bufferSerial[cursor + 5];
        if (tempType == 'x') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);

        }


      }//END 'L'

      else if (bufferSerial[0] == 'F') {
        
        int cursor = 1;
        int varSalto = 0;
        int32_t tempA = 0;
        int32_t tempB = 0;

        varSalto = (REG_FLOAT) + (4 * counterReal);

        tempB = (((int32_t)bufferSerial[cursor + 4]) << 28);
        tempA = (((int32_t)bufferSerial[cursor + 3]) << 21);
        tempB = tempB | tempA;
        tempA = (((int32_t)bufferSerial[cursor + 2]) << 14);
        tempB = tempB | tempA;
        tempA = (((int32_t)bufferSerial[cursor + 1]) << 7);
        tempB = tempB | tempA | ((int32_t)bufferSerial[cursor]);

        tempA = tempB & (0xFF);
        //EEPROM.write(varSalto + 3, tempA);
        preEEPROM[varSalto + 3] = (uint8_t)tempA;
        //delay(120);

        tempA = tempB & (0xFF00);
        tempA = tempA >> 8;
        //EEPROM.write(varSalto + 2, tempA);
        preEEPROM[varSalto + 2] = (uint8_t)tempA;
        //delay(120);

        tempA = tempB & (0xFF0000);
        tempA = tempA >> 16;
        //EEPROM.write(varSalto + 1, tempA);
        preEEPROM[varSalto + 1] = (uint8_t)tempA;
        //delay(120);

        tempA = tempB & (0xFF000000);
        tempA = tempA >> 24;
        //EEPROM.write(varSalto, tempA);
        preEEPROM[varSalto] = (uint8_t)tempA;
        //delay(120);

        counterReal++;

        int tempType = bufferSerial[cursor + 5];
        if (tempType == 'x') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);
        }


      }//end 'F'


      else if (bufferSerial[0] == '$') {
        
        bool lock = true;
        bool xFound = false;
        int i = 0;

        for(int i=0; i<(buffSize-1); i++){
          preEEPROM[countPC] = (uint8_t)bufferSerial[i];
          
          countPC++;
        }

        if(bufferSerial[buffSize-1] == 'x')
          xFound = true;

        if (xFound) {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);
        }



      }//END '$'


      else if (bufferSerial[0] == 'R') {
      
        bool lock = true;
        bool xFound = false;
        int i = 0;


        while (lock) {
          //EEPROM.write(countPC, bufferSerial[i]);
          preEEPROM[countPC] = (uint8_t)bufferSerial[i];
          //delay(120);

          countPC++;
          i++;

          if (i >= buffSize) {
            lock = false;
          }

          if (bufferSerial[i] == 'x') {
            xFound = true;
            lock = false;
          }
        }


        if (xFound) {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);
        }

      }//END 'R'


      else if (bufferSerial[0] == 'B') {
        
        bool lock = true;
        bool xFound = false;
        int i = 0;

        while (lock) {
          int salto = (REG_COUNTER + i) + (16 * countCounter);
          //EEPROM.write(salto, bufferSerial[i]);
          preEEPROM[salto] = (uint8_t)bufferSerial[i];
          //delay(120);

          i++;

          if (i >= buffSize)
            lock = false;

          if (bufferSerial[i] == 'x') {
            xFound = true;
            lock = false;
          }
        }

        if (xFound) {

          //EEPROM.write(countPC, 'B');
          preEEPROM[countPC] = (uint8_t)'B';
          //delay(120);
          countPC++;
          //EEPROM.write(countPC, (countCounter+48));
          preEEPROM[countPC] = (uint8_t)(countCounter + 48);
          //delay(120);
          countPC++;
          countCounter++;
          countCounterCTU++;

          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);
        }



      }//END 'B'


      else if (bufferSerial[0] == 'b') {
        
        bool lock = true;
        bool xFound = false;
        int i = 0;

        while (lock) {
          int salto = (REG_COUNTER + i) + (16 * countCounter);
          //EEPROM.write(salto, bufferSerial[i]);
          preEEPROM[salto] = (uint8_t)bufferSerial[i];
          //delay(120);

          i++;

          if (i >= buffSize)
            lock = false;

          if (bufferSerial[i] == 'x') {
            xFound = true;
            lock = false;
          }
        }

        if (xFound) {
          //EEPROM.write(countPC, 'b');
          preEEPROM[countPC] = (uint8_t)'b';
          //delay(120);
          countPC++;
          //EEPROM.write(countPC, (countCounter+48));
          preEEPROM[countPC] = (uint8_t)(countCounter + 48);
          //delay(120);
          countPC++;
          countCounter++;
          countCounterCTUD++;

          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);
        }

    

      }//END 'b'


      else if (bufferSerial[0] == 't') {
        
        bool lock = true;
        bool xFound = false;
        int i = 0;

        while (lock) {
          int salto = (REG_TIMER + i) + (12 * countTimer);
          //EEPROM.write(salto, bufferSerial[i]);
          preEEPROM[salto] = (uint8_t)bufferSerial[i];
          //delay(120);
          i++;

          if (i >= buffSize)
            lock = false;

          if (bufferSerial[i] == 'x') {
            xFound = true;
            lock = false;
          }
        }

        if (xFound) {

          //EEPROM.write(countPC,'q');
          preEEPROM[countPC] = (uint8_t)'q';
          //delay(120);
          countPC++;
          //EEPROM.write(countPC, (countTimer+48));
          preEEPROM[countPC] = (uint8_t)(countTimer + 48);
          //delay(120);
          countPC++;
          countTimer++;

          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);
        }


      }//END 't'


      else if (bufferSerial[0] == 'T') {
        
        bool lock = true;
        bool xFound = false;
        int i = 0;

        while (lock) {
          int salto = (REG_TIMER + i) + (12 * countTimer);
          //EEPROM.write(salto, bufferSerial[i]);
          preEEPROM[salto] = (uint8_t)bufferSerial[i];
          //delay(120);

          i++;

          if (i >= buffSize)
            lock = false;

          if (bufferSerial[i] == 'x') {
            xFound = true;
            lock = false;
          }
        }

        if (xFound) {

          //EEPROM.write(countPC,'q');
          preEEPROM[countPC] = (uint8_t)'q';
          //delay(120);
          countPC++;
          //EEPROM.write(countPC, (countTimer+48));
          preEEPROM[countPC] = (uint8_t)(countTimer + 48);
          //delay(120);
          countPC++;
          countTimer++;

          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);
        }


      }//END 'T'


      else if (bufferSerial[0] == 'S') {
        bool lock = true;
        bool xFound = false;
        int i = 0;

        while (lock) {
          int salto = (REG_TIMER + i) + (12 * countTimer);
          //EEPROM.write(salto, bufferSerial[i]);
          preEEPROM[salto] = (uint8_t)bufferSerial[i];
          //delay(120);
          i++;

          if (i >= buffSize)
            lock = false;

          if (bufferSerial[i] == 'x') {
            xFound = true;
            lock = false;
          }
        }

        if (xFound) {

          //EEPROM.write(countPC,'q');
          preEEPROM[countPC] = (uint8_t)'q';
          //delay(120);
          countPC++;
          //EEPROM.write(countPC, (countTimer+48));
          preEEPROM[countPC] = (uint8_t)(countTimer + 48);
          //delay(120);
          countPC++;
          countTimer++;

          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);
        }


      }//END 'S'


      else if (bufferSerial[0] == 'r') {
        bool lock = true;
        bool xFound = false;
        int i = 1;

        while (lock) {
          //EEPROM.write(countPC, bufferSerial[i]);
          preEEPROM[countPC] = (uint8_t)bufferSerial[i];
          //delay(120);
          countPC++;

          i++;

          if (i >= buffSize)
            lock = false;

          if (bufferSerial[i] == 'x') {
            xFound = true;
            lock = false;
          }
        }

        if (xFound) {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);

        }

      }//END 'r'


      else if (bufferSerial[0] == 'Y') {
        int cursor = 1;
        int varSalto = 0;
        int32_t tempA = 0;
        int32_t tempB = 0;

        tempB = bufferSerial[cursor + 2] << 14;
        tempA = bufferSerial[cursor + 1] << 7;
        tempB = tempB | tempA | bufferSerial[cursor];

        tempA = tempB & (0xFF);
        timeData[0] = tempA;


        int tempType = bufferSerial[cursor + 3];
        if (tempType == 'x') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);

        }

      }//END 'Y'

      else if (bufferSerial[0] == 'M') {
        int cursor = 1;
        int varSalto = 0;
        int32_t tempA = 0;
        int32_t tempB = 0;

        tempB = bufferSerial[cursor + 2] << 14;
        tempA = bufferSerial[cursor + 1] << 7;
        tempB = tempB | tempA | bufferSerial[cursor];

        tempA = tempB & (0xFF);
        timeData[1] = tempA;


        int tempType = bufferSerial[cursor + 3];
        if (tempType == 'x') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);

         
        }

      }//END 'M'


      else if (bufferSerial[0] == 'D') {
        int cursor = 1;
        int varSalto = 0;
        int32_t tempA = 0;
        int32_t tempB = 0;

        tempB = bufferSerial[cursor + 2] << 14;
        tempA = bufferSerial[cursor + 1] << 7;
        tempB = tempB | tempA | bufferSerial[cursor];

        tempA = tempB & (0xFF);
        timeData[3] = tempA;


        int tempType = bufferSerial[cursor + 3];
        if (tempType == 'x') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);

       
        }

      }//END 'D'


      else if (bufferSerial[0] == 'H') {
        int cursor = 1;
        int varSalto = 0;
        int32_t tempA = 0;
        int32_t tempB = 0;

        tempB = bufferSerial[cursor + 2] << 14;
        tempA = bufferSerial[cursor + 1] << 7;
        tempB = tempB | tempA | bufferSerial[cursor];

        tempA = tempB & (0xFF);
        timeData[4] = tempA;


        int tempType = bufferSerial[cursor + 3];
        if (tempType == 'x') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);

       
        }

      }//END 'H'


      else if (bufferSerial[0] == 'P') {
        int cursor = 1;
        int varSalto = 0;
        int32_t tempA = 0;
        int32_t tempB = 0;

        tempB = bufferSerial[cursor + 2] << 14;
        tempA = bufferSerial[cursor + 1] << 7;
        tempB = tempB | tempA | bufferSerial[cursor];

        tempA = tempB & (0xFF);
        timeData[5] = tempA;


        int tempType = bufferSerial[cursor + 3];
        if (tempType == 'x') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);

        }

      }//END 'P'


      else if (bufferSerial[0] == 'k') {
        int cursor = 1;
        int varSalto = 0;
        int32_t tempA = 0;
        int32_t tempB = 0;

        tempB = bufferSerial[cursor + 2] << 14;
        tempA = bufferSerial[cursor + 1] << 7;
        tempB = tempB | tempA | bufferSerial[cursor];

        tempA = tempB & (0xFF);
        timeData[6] = tempA;


        int tempType = bufferSerial[cursor + 3];
        if (tempType == 'x') {
          clientTemp.flush();
          while (clientTemp.read() != -1) {

          }

          clientTemp.write((uint8_t)'y');
          clientTemp.write((uint8_t)10);

          updateMyTime = true;
        }

      }//END 'k'


      else if (bufferSerial[0] == 'd') {
                                                 
      
        preEEPROM[countPC] = (uint8_t)'z';
        
        preEEPROM[3959] = (uint8_t)'2';
    

        clientTemp.stop();
        clientConnected = false;

        cleanOldCounters(countCounter);
        cleanOldTimers(countTimer);
       
        commitToEEPROM();


        if (updateMyTime && (!rtcUpdated)) {
          updateMyTime = false;
          updateAllTime();
        }


        programming = false;   
        stopProcess();

        initProgram();
        
        drawBackground();
        preDraw();
                                                         
      }


      else if (bufferSerial[0] == 'C') {                                                       //RTC UPDATE
                                                    
        clientTemp.write((uint8_t)'t');
        clientTemp.write((uint8_t)10);
        delay(10);

        int32_t tempA = 0;
        int32_t tempB = 0;
        bool updatingTime = true;


        while (updatingTime) {
          if (clientTemp.available() > 0) {

            buffSize = clientTemp.readBytesUntil(42, bufferSerial, sizeof(bufferSerial));   //42 *
            //buffSize = readMyByteUntil(42, clientTemp);

            if (buffSize > 0) {
              switch (bufferSerial[0]) {

                case'x':
                  clientTemp.flush();
                  while (clientTemp.read() != -1) {

                  }
                  //Serial.println('y');
                  clientTemp.write((uint8_t)'p');
                  clientTemp.write((uint8_t)10);

                  break;

                case 'Y':
                  tempB = bufferSerial[3] << 14;
                  tempA = bufferSerial[2] << 7;
                  tempB = tempB | tempA | bufferSerial[1];

                  tempA = tempB & (0xFF);
                  timeData[0] = tempA;

                  break;

                case 'M':
                  tempB = bufferSerial[3] << 14;
                  tempA = bufferSerial[2] << 7;
                  tempB = tempB | tempA | bufferSerial[1];

                  tempA = tempB & (0xFF);
                  timeData[1] = tempA;

                  break;

                case 'D':
                  tempB = bufferSerial[3] << 14;
                  tempA = bufferSerial[2] << 7;
                  tempB = tempB | tempA | bufferSerial[1];

                  tempA = tempB & (0xFF);
                  timeData[3] = tempA;

                  break;

                case 'H':
                  tempB = bufferSerial[3] << 14;
                  tempA = bufferSerial[2] << 7;
                  tempB = tempB | tempA | bufferSerial[1];

                  tempA = tempB & (0xFF);
                  timeData[4] = tempA;

                  break;

                case 'P':
                  tempB = bufferSerial[3] << 14;
                  tempA = bufferSerial[2] << 7;
                  tempB = tempB | tempA | bufferSerial[1];

                  tempA = tempB & (0xFF);
                  timeData[5] = tempA;

                  break;

                case 'k':
                  tempB = bufferSerial[3] << 14;
                  tempA = bufferSerial[2] << 7;
                  tempB = tempB | tempA | bufferSerial[1];

                  tempA = tempB & (0xFF);
                  timeData[6] = tempA;


                  break;

                case 'd':

                  updatingTime = false;

                  clientTemp.flush();
                  while (clientTemp.read() != -1) {

                  }

                  break;

              }//END switch
            }//END if
          }//END if

        }//END while

        updateAllTime();

        clientTemp.stop();
        clientConnected = false;

                                                           
      }//END else if(bufferSerial[0] == 'C')

    }//END if(inReprogramTime)

    else {                                                  //HMI mode
                                                             
      int buffSize = 0;

      buffSize = readMyByteUntil(98, clientTemp);

      if (buffSize > 0) {

        if (!configStream) {

          if ((bufferSerial[0] == 'c')) {
            sendDataConfig = false;
            clientTemp.write((uint8_t)100);
            clientTemp.write((uint8_t)'s');

            int tempSizeI = 0;
            int tempSizeO = 0;

            for (int i = 0; i < 3; i++) { //Input  
              if (iMask[i] == true)
                tempSizeI++;
            }

            for (int i = 0; i < 4; i++) { //Output
              if (oMask[i] == true)
                tempSizeO++;
            }

            fragmentData(tempSizeI, false, false);
            clientTemp.write((uint8_t)md);
            clientTemp.write((uint8_t)ld);

            clientTemp.write((uint8_t)'i');
            if (tempSizeI > 0) {
              for (int i = 0; i < 3; i++) { //input
                if (iMask[i] == true)
                  clientTemp.write((uint8_t)i);
              }
            }

            fragmentData(tempSizeO, false, false);
            clientTemp.write((uint8_t)md);
            clientTemp.write((uint8_t)ld);

            clientTemp.write((uint8_t)'o');

            if (tempSizeO > 0) {
              for (int i = 0; i < 4; i++) { //output
                if (oMask[i] == true)
                  clientTemp.write((uint8_t)i);
              }
            }

            tempSizeI = 0;
            for (int i = 0; i < 64; i++) {
              if (vMask[i] == true)
                tempSizeI++;
            }

            fragmentData(tempSizeI, false, false);
            clientTemp.write((uint8_t)md);
            clientTemp.write((uint8_t)ld);

            clientTemp.write((uint8_t)'v');

            if (tempSizeI > 0) {
              for (int i = 0; i < 64; i++) {
                if (vMask[i] == true) {
                  fragmentData(i, false, false);
                  clientTemp.write((uint8_t)md);
                  clientTemp.write((uint8_t)ld);
                }
              }
            }

            fragmentData((uint16_t)EEPROM.read(4057), false, false);
            clientTemp.write((uint8_t)md);
            clientTemp.write((uint8_t)ld);

            fragmentData((uint16_t)EEPROM.read(4058), false, false);
            clientTemp.write((uint8_t)md);
            clientTemp.write((uint8_t)ld);

            fragmentData((uint16_t)EEPROM.read(4059), false, false);
            clientTemp.write((uint8_t)md);
            clientTemp.write((uint8_t)ld);

            clientTemp.write((uint8_t)98);
            clientTemp.flush();

            //fragmentData()
          }

          else if (bufferSerial[0] == 'f') {
            //clientTemp.flush();
            configStream = true;
            sendStartSend = true;
          }


        }//if(!configStream)

        else {
          if ((bufferSerial[0] == 'w')) {
            inProcessLock = true;
            bool oOk = false, vOk = false, eOk = false, dOk = false, floatOk = false;
            int cursor = 1;
            int newCursor = 0;
            int dataIndex = 0;

            int tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            int tempType = bufferSerial[cursor + 2];
            cursor += 3;


            if (tempType == 'o') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  dataIndex = bufferSerial[cursor + (i * 2)];
                  if (bufferSerial[cursor + (i * 2) + 1] == 1) {
                    oState[dataIndex] = true;
                  }
                  else {
                    oState[dataIndex] = false;
                  }

                  newCursor = cursor + (i * 2) + 2;

                }
                cursor = newCursor;
              }
              oOk = true;
            }

            tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            tempType = bufferSerial[cursor + 2];
            cursor += 3;

            if (tempType == 'v') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  dataIndex = bufferSerial[cursor + (i * 2)];

                  if (bufferSerial[cursor + (i * 2) + 1] == 1) {
                    vState[dataIndex] = true;
                  }
                  else {
                    vState[dataIndex] = false;
                  }

                  newCursor = cursor + (i * 2) + 2;

                }
                cursor = newCursor;
              }
              vOk = true;
            }


            tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            tempType = bufferSerial[cursor + 2];
            cursor += 3;

            if (tempType == 'e') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  dataIndex = bufferSerial[cursor + (i * 4)];

                  int tempValue2 = joinInt16(bufferSerial[cursor + (i * 4) + 1], bufferSerial[cursor + (i * 4) + 2], bufferSerial[cursor + (i * 4) + 3]);
                  intValues[dataIndex] = tempValue2;

                  newCursor = cursor + (i * 4) + 4;

                }
                cursor = newCursor;
              }
              eOk = true;
            }

            tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            tempType = bufferSerial[cursor + 2];
            cursor += 3;

            if (tempType == 'd') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  dataIndex = bufferSerial[cursor + (i * 7)];

                  int tempValue2 = joinInt32(bufferSerial[cursor + (i * 7) + 1], bufferSerial[cursor + (i * 7) + 2], bufferSerial[cursor + (i * 7) + 3], bufferSerial[cursor + (i * 7) + 4], bufferSerial[cursor + (i * 7) + 5], bufferSerial[cursor + (i * 7) + 6]);
                  dintValues[dataIndex] = tempValue2;

                  newCursor = cursor + (i * 7) + 7;

                }
                cursor = newCursor;
              }
              dOk = true;
            }

            tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            tempType = bufferSerial[cursor + 2];
            cursor += 3;

            if (tempType == 'f') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  dataIndex = bufferSerial[cursor + (i * 7)];

                  uint32_t tempUint32 = joinInt32(bufferSerial[cursor + (i * 7) + 1], bufferSerial[cursor + (i * 7) + 2], bufferSerial[cursor + (i * 7) + 3], bufferSerial[cursor + (i * 7) + 4], bufferSerial[cursor + (i * 7) + 5], bufferSerial[cursor + (i * 7) + 6]);
                  float tempFloat = joinFloat(tempUint32);
                  realValues[dataIndex] = tempFloat;

                  newCursor = cursor + (i * 7) + 7;

                }
                cursor = newCursor;
              }
              floatOk = true;
            }

            if (oOk && vOk && eOk && dOk && floatOk) {
              requestSecondPart = true;
            }
            else {
              inProcessLock = false;
            }

          }//END if((bufferSerial[0]=='w'))

          else if (bufferSerial[0] == 'y') {
            inProcessLock = true;
            bool iOk = false, oOk = false, vOk = false, eOk = false, dOk = false, floatOk = false;

            int cursor = 1;
            int newCursor = 0;
            int dataIndex = 0;

            int tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            int tempType = bufferSerial[cursor + 2];
            cursor += 3;

            iSize = tempValue;

            if (tempType == 'i') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  iIndex[i] = bufferSerial[cursor + i];

                  newCursor = cursor + i + 1;
                }
                cursor = newCursor;
              }
              iOk = true;
            }

            tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            tempType = bufferSerial[cursor + 2];
            cursor += 3;

            oSize = tempValue;

            if (tempType == 'o') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  dataIndex = bufferSerial[cursor + i];
                  oIndex[i] = dataIndex;

                  newCursor = cursor + i + 1;
                }
                cursor = newCursor;
              }
              oOk = true;
            }


            tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            tempType = bufferSerial[cursor + 2];
            cursor += 3;

            vSize = tempValue;

            if (tempType == 'v') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  dataIndex = bufferSerial[cursor + i];
                  vIndex[i] = dataIndex;

                  newCursor = cursor + i + 1;

                }
                cursor = newCursor;
              }
              vOk = true;
            }


            tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            tempType = bufferSerial[cursor + 2];
            cursor += 3;

            eSize = tempValue;

            if (tempType == 'e') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  dataIndex = bufferSerial[cursor + (i)];
                  intIndex[i] = dataIndex;

                  newCursor = cursor + i + 1;

                }
                cursor = newCursor;
              }
              eOk = true;
            }

            tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            tempType = bufferSerial[cursor + 2];
            cursor += 3;

            dSize = tempValue;

            if (tempType == 'd') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  dataIndex = bufferSerial[cursor + i];
                  dintIndex[i] = dataIndex;

                  newCursor = cursor + i + 1;

                }
                cursor = newCursor;
              }
              dOk = true;
            }

            tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
            tempType = bufferSerial[cursor + 2];
            cursor += 3;

            floatSize = tempValue;

            if (tempType == 'f') {
              if (tempValue > 0) {
                for (int i = 0; i < tempValue; i++) {
                  dataIndex = bufferSerial[cursor + i];
                  realIndex[i] = dataIndex;

                  newCursor = cursor + i + 1;

                }
                cursor = newCursor;
              }
              floatOk = true;
            }

            cleanClientData(server);

            if (!clientTemp.connected()){
                                                               
            }

            if (iOk && oOk && vOk && eOk && dOk && floatOk) {
              clientTemp.write((uint8_t)'z');
              clientTemp.write((uint8_t)'i');

              if (iSize > 0) {
                for (int i = 0; i < iSize; i++) {
                  dataIndex = iIndex[i];
                  if (iState[dataIndex])
                    clientTemp.write((uint8_t)2);
                  else
                    clientTemp.write((uint8_t)1);
                }//END for
              }//END if(iSize > 0)

              clientTemp.write((uint8_t)'o');

              if (oSize > 0) {
                for (int i = 0; i < oSize; i++) {
                  dataIndex = oIndex[i];
                  if (oState[dataIndex])
                    clientTemp.write((uint8_t)2);
                  else
                    clientTemp.write((uint8_t)1);
                }//END for
              }//END if

              clientTemp.write((uint8_t)'v');

              if (vSize > 0) {
                for (int i = 0; i < vSize; i++) {
                  dataIndex = vIndex[i];
                  if (vState[dataIndex])
                    clientTemp.write((uint8_t)2);
                  else
                    clientTemp.write((uint8_t)1);
                }//END for
              }//END if

              clientTemp.write((uint8_t)'e');

              if (eSize > 0) {
                for (int i = 0; i < eSize; i++) {
                  dataIndex = intIndex[i];

                  int dataTemp = intValues[dataIndex];
                  fragmentData(dataTemp);
                  clientTemp.write((uint8_t)hd);
                  clientTemp.write((uint8_t)md);
                  clientTemp.write((uint8_t)ld);
                }//END for
              }//END if

              clientTemp.write((uint8_t)'d');

              if (dSize > 0) {
                for (int i = 0; i < dSize; i++) {
                  dataIndex = dintIndex[i];

                  int dataTemp = dintValues[dataIndex];
                  fragmentData(dataTemp, (uint8_t)4);
                  clientTemp.write((uint8_t)hd32);
                  clientTemp.write((uint8_t)md32);
                  clientTemp.write((uint8_t)ld32);
                  clientTemp.write((uint8_t)hd);
                  clientTemp.write((uint8_t)md);
                  clientTemp.write((uint8_t)ld);
                }//END for
              }//END if

              clientTemp.write((uint8_t)'f');

              if (floatSize > 0) {
                for (int i = 0; i < floatSize; i++) {
                  dataIndex = realIndex[i];

                  floatValue tempFloat;
                  tempFloat.val = realValues[dataIndex];

                  uint32_t tempUint32 = (uint32_t)tempFloat.bytes[0];
                  tempUint32 |= ((uint32_t)tempFloat.bytes[1] << 8);
                  tempUint32 |= ((uint32_t)tempFloat.bytes[2] << 16);
                  tempUint32 |= ((uint32_t)tempFloat.bytes[3] << 24);

                  fragmentData(tempUint32, (uint8_t)4);
                  clientTemp.write((uint8_t)hd32);
                  clientTemp.write((uint8_t)md32);
                  clientTemp.write((uint8_t)ld32);
                  clientTemp.write((uint8_t)hd);
                  clientTemp.write((uint8_t)md);
                  clientTemp.write((uint8_t)ld);

                }//END for
              }//END if

              clientTemp.write((uint8_t)'b');
            }

            delay(150);

            alreadySendIt = false;
            sendStartSend = true;
            inProcessLock = false;

          }//END else if(bufferSerial[0] == 'y')


        }//else of if(!configStream)

      }//END if(buffSize > 0)
                                                            
    } //END else HMI mode

  }//END if(clientTemp.available())

  if (runReprogramTime) {

    unsigned long currentTimeTemp = millis();
    currentTimeTemp -= startReprogramTime;


    if (currentTimeTemp > targetReprogramTime) {
      inReprogramTime = false;
      runReprogramTime = false;
                                                              
    }


  }

  if (sendStartSend) {
    cleanClientData(server);

    clientTemp.write((uint8_t)'w');
    clientTemp.write((uint8_t)'b');

    sendStartSend = false;
    alreadySendIt = true;
    inSecondState = false;
    inProcessLock = false;
    startTimeSend = millis();

  }//END if(sendStartSend)


  if (requestSecondPart) {
    cleanClientData(server);

    clientTemp.write((uint8_t)'y');
    clientTemp.write((uint8_t)'b');
    requestSecondPart = false;
    inSecondState = true;

    startTimeSend = millis();
    inProcessLock = false;
  }//END if(requestSecondPart)


  if (alreadySendIt) {
    if (!inProcessLock) {
      unsigned long tempCurrent = millis();
      tempCurrent -= startTimeSend;
      if (tempCurrent > targetTimeSend) {
        Serial.print("Re-connecting: ");
        Serial.println(clientTemp.connected());

        if (!clientTemp.connected()){
                                                           
        }

        cleanClientData(server);

        clientTemp.write((uint8_t)'w');
        clientTemp.write((uint8_t)'b');

        sendStartSend = false;
        alreadySendIt = true;
        inSecondState = false;
        inProcessLock = false;
        startTimeSend = millis();

      }//END if(tempCurrent > targetTimeSend)
    }//END if(!inProcessLock)
  }//END if(alreadySendIt)



}//END reprogram


void cleanOldCounters(int tempCountCounter) {
  if (tempCountCounter < 25) {
    for (int i = tempCountCounter; i < 25; i++) {
      int salto = (3574) + (16 * i);
      //EEPROM.write(salto, '~');
      preEEPROM[salto] = (uint8_t)'~';
      //delay(120);

    }
  }
}//END cleanOldCounters


void cleanOldTimers(int tempCountTimer) {
  if (tempCountTimer < 25) {
    for (int i = tempCountTimer; i < 25; i++) {
      int salto = (3285) + (12 * i);
      //EEPROM.write(salto, '~');
      preEEPROM[salto] = (uint8_t)'~';
      //delay(120);

    }
  }
}//END cleanOldCounters


bool setDnCounter(int posi, bool dnState) {
  bool tempReturn = dnState;
  if (!dnState) {
    if (dintValues[myCounter[posi].countIndex] >= dintValues[myCounter[posi].presetIndex]) {
      tempReturn = true;
    }//END if(myCounter[posi].count >= myCounter[posi].preset)
  }//END if
  else {
    if (myCounter[posi].resetNo) {
      tempReturn = false;
    }
  }
  return tempReturn;
}

bool setDnCounterCTUD(int posi, bool dnState) {
  bool tempReturn = dnState;
  if (!dnState) {
    if (dintValues[myCounterCTUD[posi].countIndex] >= dintValues[myCounterCTUD[posi].presetIndex]) {
      tempReturn = true;
    }//END if(myCounterCTUD[posi].count >= myCounterCTUD[posi].preset)
  }//END if
  else {
    if (myCounterCTUD[posi].resetNo) {
      tempReturn = false;
    }

    if (dintValues[myCounterCTUD[posi].countIndex] < dintValues[myCounterCTUD[posi].presetIndex]) {
      tempReturn = false;
    }//END if(myCounterCTUD[posi].count < myCounterCTUD[posi].preset)
  }
  return tempReturn;
}



int readAnalogPin(int index) {

  addAnalogInputIndex(index);

  analogInputs[index] = analogRead(iPins[index]);
  
  return analogInputs[index];
}

void writeServoPin(int index, int value) {   

  if (!servoOutput[index]) {
    servoOutput[index] = true;

    switch (index) {
      case 0:
        ledcAttachPin(oPins[index], 8);
      break;
      case 1:
        ledcAttachPin(oPins[index], 9);
      break;
      case 2:
        ledcAttachPin(oPins[index], 10);
      break;
      case 3:
        ledcAttachPin(oPins[index], 11);
      break;
      case 4:
        ledcAttachPin(oPins[index], 12);
      break;
      case 5:
        ledcAttachPin(oPins[index], 13);
      break;
      case 6:
        ledcAttachPin(oPins[index], 14);
      break;
      case 7:
        ledcAttachPin(oPins[index], 15);
      break;
    }

  }

  if (value > 180)
    value = 180;
  if (value < 0)
    value = 0;

  switch (index) {
    case 0:
        writeServo(8, value);
      break;
      case 1:
        writeServo(9, value);
      break;
      case 2:
        writeServo(10, value);
      break;
      case 3:
        writeServo(11, value);
      break;
      case 4:
        writeServo(12, value);
      break;
      case 5:
        writeServo(13, value);
      break;
      case 6:
        writeServo(14, value);
      break;
      case 7:
        writeServo(15, value);
      break;

  }
  delay(15);

  if (value == 0)
    value = 1;

  analogOutputs[index] = value;
}

void writeAnalogPin(int index, int value) {    

  if (!pwmOutput[index]) {
    pwmOutput[index] = true;

    switch (index) {
      case 0:
        ledcAttachPin(oPins[index], 0);
      break;
      case 1:
        ledcAttachPin(oPins[index], 1);
      break;
      case 2:
        ledcAttachPin(oPins[index], 2);
      break;
      case 3:
        ledcAttachPin(oPins[index], 3);
      break;
      case 4:
        ledcAttachPin(oPins[index], 4);
      break;
      case 5:
        ledcAttachPin(oPins[index], 5);
      break;
      case 6:
        ledcAttachPin(oPins[index], 6);
      break;
      case 7:
        ledcAttachPin(oPins[index], 7);
      break;
    }

  }
  

  if (value > 4095)
    value = 4095;
  if (value < 0)
    value = 0;

  switch (index) {
    case 0:
        ledcWrite(0, value);
      break;
      case 1:
        ledcWrite(1, value);
      break;
      case 2:
        ledcWrite(2, value);
      break;
      case 3:
        ledcWrite(3, value);
      break;
      case 4:
        ledcWrite(4, value);
      break;
      case 5:
        ledcWrite(5, value);
      break;
      case 6:
        ledcWrite(6, value);
      break;
      case 7:
        ledcWrite(7, value);
      break;

  }

  analogOutputs[index] = value;
  delay(20);
}

int readAnalogOutput(int index) {
  return analogOutputs[index];
}

void equ() {                                                                    //Equal
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC += 2;
  tempSource = getSource();
  PC++;

  //Serial.print("SourceA: ");
  //Serial.println(tempSource);

  tagIsEnable = isTagEnable(PC + 3);

  if (tagIsEnable) {
    PC += 3;
    tempSourceB = getSource();
  }
  else {
    tempSourceB = (float)getJoinValue();
    PC += 4;
  }

  //Serial.print("SourceB: ");
  //Serial.println(tempSourceB);

  if (tempSource == tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);
}

void neq() {                                                                    //Not equal
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC += 2;
  tempSource = getSource();
  PC++;

  tagIsEnable = isTagEnable(PC + 3);

  if (tagIsEnable) {
    PC += 3;
    tempSourceB = getSource();
  }
  else {
    tempSourceB = (float)getJoinValue();
    PC += 4;
  }


  if (tempSource != tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);
}

void les() {                                                                    //Less than
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC += 2;
  tempSource = getSource();
  PC++;

  tagIsEnable = isTagEnable(PC + 3);

  if (tagIsEnable) {
    PC += 3;
    tempSourceB = getSource();
  }
  else {
    tempSourceB = (float)getJoinValue();
    PC += 4;
  }


  if (tempSource < tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);

  //Serial.println("les");
}

void leq() {                                                                    //Less and equal
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC += 2;
  tempSource = getSource();
  PC++;

  tagIsEnable = isTagEnable(PC + 3);

  if (tagIsEnable) {
    PC += 3;
    tempSourceB = getSource();
  }
  else {
    tempSourceB = (float)getJoinValue();
    PC += 4;
  }


  if (tempSource <= tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);
}

void grt() {                                                                    //Greater than
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC += 2;
  tempSource = getSource();
  PC++;

  tagIsEnable = isTagEnable(PC + 3);

  if (tagIsEnable) {
    PC += 3;
    tempSourceB = getSource();
  }
  else {
    tempSourceB = (float)getJoinValue();
    PC += 4;
  }


  if (tempSource > tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);

  //Serial.println("GRT");
}

void geq() {                                                                    //Greater and equal
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC += 2;
  tempSource = getSource();
  PC++;

  tagIsEnable = isTagEnable(PC + 3);

  if (tagIsEnable) {
    PC += 3;
    tempSourceB = getSource();
  }
  else {
    tempSourceB = (float)getJoinValue();
    PC += 4;
  }


  if (tempSource >= tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);
}


void mov() {                                                                    //Move
  bool noState = false;

  PC++;

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;
  }

  if (noState) {
    float tempSource = 0;

    PC++;
    if (isTagEnable(PC)) {
      tempSource = getSource();
      PC += 4;
    }
    else {
      PC += 2;
      tempSource = (float)getJoinValue();
      PC += 3;
    }

    //Serial.print("Mov: ");                                         
    //Serial.println(tempSource);
    //Serial.println("");
    setDest(tempSource);
  }
  else
    PC += 7;


}

void lim() {                                                                    //Limit
  int dnIndex = 0;
  float tempTest = 0;
  float tempSourceMax = 0;
  float tempSourceMin = 0;
  bool tagIsEnable = false;

  PC++;                                                                         //byte 2
  dnIndex = PC;
  PC += 2;                                                                      //byte 4
  tempTest = getSource();
  PC++;                                                                         //byte 6
  tagIsEnable = isTagEnable(PC + 3);

  if (tagIsEnable) {
    PC += 3;                                                                    //byte 9
    tempSourceMin = getSource();

  }
  else {
    tempSourceMin = (float)getJoinValue();
    PC += 4;                                                                    //byte 10
  }
  PC++;                                                                         //byte 11

  tagIsEnable = isTagEnable(PC + 3);

  if (tagIsEnable) {
    PC += 3;                                                                    //byte 14
    tempSourceMax = getSource();
  }
  else {
    tempSourceMax = (float)getJoinValue();
    PC += 4;
  }


  if ((tempTest >= tempSourceMin) && (tempTest <= tempSourceMax))
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);

  //Serial.println("lim");

}

void meq() {                                                                    //Mask equal
  int dnIndex = 0;
  int32_t tempSource = 0;
  int32_t tempMask = 0;
  int32_t tempCompare = 0;
  bool tagIsEnable = false;

  PC++;                                                                         //byte 2
  dnIndex = PC;
  PC += 2;                                                                      //byte 4
  tempSource = getSource();
  PC++;                                                                         //byte 6
  tagIsEnable = isTagEnable(PC + 3);

  if (tagIsEnable) {
    PC += 3;                                                                    //byte 9
    tempMask = getSource();

  }
  else {
    tempMask = (float)getJoinValue();
    PC += 4;                                                                    //byte 10
  }
  PC++;                                                                         //byte 11

  tagIsEnable = isTagEnable(PC + 3);

  if (tagIsEnable) {
    PC += 3;                                                                    //byte 14
    tempCompare = getSource();
  }
  else {
    tempCompare = (float)getJoinValue();
    PC += 4;
  }

  tempSource = tempSource & tempMask;

  if (tempSource == tempCompare)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);
}//END meq()

void andOp() {                                                                  //And
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    int32_t tempSourceA = 0;
    int32_t tempSourceB = 0;

    PC++;                                                                     //byte 4
    tempSourceA = (int32_t)getSource();
    PC++;                                                                     //byte 6

    if (isTagEnable(PC + 3)) {                                                //(byte 9)
      PC += 3;                                                                //byte 9
      tempSourceB = (int32_t)getSource();
      PC++;                                                                   //byte 11
    }
    else {
      tempSourceB = (int32_t)getJoinValue();
      PC += 5;                                                                //byte 11
    }

    tempSourceA = tempSourceA & tempSourceB;

    setDest(tempSourceA);
  }
  else
    PC += 9;                                                                    //byte 12

}//END andOp()

void orOp() {                                                                   //Or
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    int32_t tempSourceA = 0;
    int32_t tempSourceB = 0;

    PC++;                                                                     //byte 4
    tempSourceA = (int32_t)getSource();
    PC++;                                                                     //byte 6

    if (isTagEnable(PC + 3)) {                                                //(byte 9)
      PC += 3;                                                                //byte 9
      tempSourceB = (int32_t)getSource();
      PC++;                                                                   //byte 11
    }
    else {
      tempSourceB = (int32_t)getJoinValue();
      PC += 5;                                                                //byte 11
    }

    tempSourceA = tempSourceA | tempSourceB;

    setDest(tempSourceA);
  }
  else
    PC += 9;
}//END orOp()

void notOp() {                                                                  //Not
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    int32_t tempSource32 = 0;
    int16_t tempSource16 = 0;
    float tempSource = 0;

    PC++;                                                                     //byte 4

    if (isSourceInt()) {
      tempSource16 = (int16_t)getSource();
      tempSource16 = ~(tempSource16);
      tempSource = (float)tempSource16;
    }
    else {
      tempSource32 = (int32_t)getSource();
      tempSource32 = ~(tempSource32);
      tempSource = (float)tempSource32;
    }


    PC++;                                                                     //byte 6

    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}//END notOp()

void xorOp() {                                                                  //Xor
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    int32_t tempSourceA = 0;
    int32_t tempSourceB = 0;

    PC++;                                                                     //byte 4
    tempSourceA = (int32_t)getSource();
    PC++;                                                                     //byte 6

    if (isTagEnable(PC + 3)) {                                                //(byte 9)
      PC += 3;                                                                //byte 9
      tempSourceB = (int32_t)getSource();
      PC++;                                                                   //byte 11
    }
    else {
      tempSourceB = (int32_t)getJoinValue();
      PC += 5;                                                                //byte 11
    }

    tempSourceA = tempSourceA ^ tempSourceB;

    setDest(tempSourceA);
  }
  else
    PC += 9;
}//END xorOp()

void addOp() {                                                                  //Add
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    float tempSourceA = 0;
    float tempSourceB = 0;

    PC++;                                                                     //byte 4
    tempSourceA = getSource();
    PC++;                                                                     //byte 6

    if (isTagEnable(PC + 3)) {                                                //(byte 9)
      PC += 3;                                                                //byte 9
      tempSourceB = getSource();
      PC++;                                                                   //byte 11
    }
    else {
      tempSourceB = (float)getJoinValue();
      PC += 5;                                                                //byte 11
    }

    tempSourceA += tempSourceB;

    setDest(tempSourceA);
    //Serial.println("ADD");
  }
  else
    PC += 9;
}

void subOp() {                                                                  //Sub
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    float tempSourceA = 0;
    float tempSourceB = 0;

    PC++;                                                                     //byte 4
    tempSourceA = getSource();
    PC++;                                                                     //byte 6

    if (isTagEnable(PC + 3)) {                                                //(byte 9)
      PC += 3;                                                                //byte 9
      tempSourceB = getSource();
      PC++;                                                                   //byte 11
    }
    else {
      tempSourceB = (float)getJoinValue();
      PC += 5;                                                                //byte 11
    }

    tempSourceA -= tempSourceB;

    setDest(tempSourceA);
  }
  else
    PC += 9;
}

void mulOp() {                                                                  //Mul
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSourceA = 0;
    float tempSourceB = 0;

    PC++;                                                                     //byte 4
    tempSourceA = getSource();
    PC++;                                                                     //byte 6

    if (isTagEnable(PC + 3)) {                                                //(byte 9)

      PC += 3;                                                                //byte 9
      tempSourceB = getSource();

      PC++;                                                                   //byte 11
    }
    else {
      tempSourceB = (float)getJoinValue();

      PC += 5;                                                                //byte 11
    }


    tempSourceA *= tempSourceB;

    setDest(tempSourceA);
    //Serial.println("MUL");
  }
  else
    PC += 9;
}

void divOp() {                                                                  //Div
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    float tempSourceA = 0;
    float tempSourceB = 0;

    PC++;                                                                     //byte 4
    tempSourceA = getSource();
    PC++;                                                                     //byte 6

    if (isTagEnable(PC + 3)) {                                                //(byte 9)
      PC += 3;                                                                //byte 9
      tempSourceB = getSource();
      PC++;                                                                   //byte 11
    }
    else {
      tempSourceB = (float)getJoinValue();
      PC += 5;                                                                //byte 11
    }

    if (tempSourceB != 0)
      tempSourceA /= tempSourceB;

    setDest(tempSourceA);
  }
  else
    PC += 9;
}

void xpy() {                                                                    //x pow y
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSourceA = 0;
    float tempSourceB = 0;

    PC++;                                                                     //byte 4
    tempSourceA = getSource();
    PC++;                                                                     //byte 6

    if (isTagEnable(PC + 3)) {                                                //(byte 9)
      PC += 3;                                                                //byte 9
      tempSourceB = getSource();
      PC++;                                                                   //byte 11
    }
    else {
      tempSourceB = (float)getJoinValue();
      PC += 5;                                                                //byte 11
    }

    tempSourceA = pow(tempSourceA, tempSourceB);

    setDest(tempSourceA);
  }
  else
    PC += 9;
}

void absMath() {                                                                //Abs
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSource = 0;

    PC++;                                                                     //byte 4
    tempSource = getSource();
    PC++;                                                                     //byte 6

    tempSource = abs(tempSource);

    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}

void sqrMath() {
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSource = 0;

    PC++;                                                                     //byte 4
    tempSource = getSource();
    PC++;                                                                     //byte 6

    tempSource = sqrt(tempSource);

    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}

void asnMath() {
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSource = 0;

    PC++;                                                                     //byte 4
    tempSource = getSource();
    PC++;                                                                     //byte 6

    tempSource = asin(tempSource);

    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}

void acsMath() {
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSource = 0;

    PC++;                                                                     //byte 4
    tempSource = getSource();
    PC++;                                                                     //byte 6

    tempSource = acos(tempSource);

    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}

void atnMath() {
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSource = 0;

    PC++;                                                                     //byte 4
    tempSource = getSource();
    PC++;                                                                     //byte 6

    tempSource = atan(tempSource);

    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}

void cosMath() {
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSource = 0;

    PC++;                                                                     //byte 4
    tempSource = getSource();
    PC++;                                                                     //byte 6

    tempSource = cos(tempSource);

    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}

void lnMath() {
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSource = 0;

    PC++;                                                                     //byte 4
    tempSource = getSource();
    PC++;                                                                     //byte 6

    tempSource = log(tempSource);

    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}

void logMath() {
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSource = 0;

    PC++;                                                                     //byte 4
    tempSource = getSource();
    PC++;                                                                     //byte 6

    tempSource = log10(tempSource);

    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}

void sinMath() {
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSource = 0;

    PC++;                                                                     //byte 4
    tempSource = getSource();
    PC++;                                                                     //byte 6

    tempSource = sin(tempSource);

    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}

void tanMath() {
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }

  if (noState) {
    float tempSource = 0;

    PC++;                                                                     //byte 4
    tempSource = getSource();
    PC++;                                                                     //byte 6

    tempSource = tan(tempSource);
    //Serial.print("tan: ");
    //Serial.println(tempSource);
    //Serial.println("");
    setDest(tempSource);
  }
  else
    PC += 4;                                                                    //byte 7
}

void bsl() {                                                                    //Bit Shift Left
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC)) {
    bool tempPre = getUpEdgeActive();
    bool noStateShift = getNoValue();

    if (tempPre)
      noState = true;

    if (!noStateShift) {
      if (isTagEnable(PC + 1))
        setDnValue(false, PC + 1);
    }

  }
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    int dnIndex = 0;
    bool dnEnabled = false;
    int ulIndex = 0;
    bool ulEnabled = false;
    int32_t tempInt = 0;
    int sourceIndex = 0;
    bool bitSource = false;
    int32_t tempVarLength = 0;
    bool ulBit = false;

    PC++;                                                                       //byte 4
    if (isTagEnable(PC)) {                                                      //if dn is enabled
      dnEnabled = true;
      dnIndex = PC;
    }

    PC += 2;                                                                    //byte 6

    if (isTagEnable(PC)) {
      ulEnabled = true;
      ulIndex = PC;
    }

    PC += 2;                                                                    //byte 8

    bool typeVar = isTagEnable(PC);
    PC++;                                                                       //byte 9
    sourceIndex = PC;

    if (typeVar) {                                                              //int-dint
      int32_t tempUl = 0;

      tempInt = (int32_t)getSource();
      PC++;                                                                     //byte 11

      if (isTagEnable(PC + 1)) {
        PC++;                                                                   //byte 12
        bitSource = getNoValue();
        PC++;                                                                   //byte 14
      }
      else {
        if (1 == (int8_t)EEPROM.read(PC))
          bitSource = true;
        PC += 3;                                                                //byte 14
      }

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 17
        tempVarLength = getSource();
      }
      else {
        tempVarLength = (int32_t)getJoinValue();
        PC += 4;                                                                //byte 18
      }

      //Logic
      if (ulEnabled) {
        tempUl = (tempInt & ((int32_t)1 << (tempVarLength - 1)));

        tempUl = (int32_t)(tempUl >> (tempVarLength - 1));
        if (tempUl == 1)
          ulBit = true;
        else
          ulBit = false;
      }

      int32_t maskUl = 1;
      for (int j = 0; j < (tempVarLength - 1); j++) {
        maskUl = (maskUl << 1);
        maskUl = (maskUl | 1);
      }

      tempUl = tempInt;

      tempUl = tempUl & maskUl;
      tempUl = tempUl << 1;
      tempUl = tempUl & maskUl;

      if (bitSource)
        tempUl = tempUl | 1;

      int32_t mask2 = ~maskUl;

      tempInt = (tempInt & mask2);

      tempInt = tempInt | tempUl;

      updateVariable(sourceIndex, tempInt);

    }//END int-dint

    else {                                                                      //output
      int outIndex = getOutIndex();
      PC++;                                                                     //byte 11

      if (isTagEnable(PC + 1)) {
        PC++;                                                                   //byte 12
        bitSource = getNoValue();
        PC++;                                                                   //byte 14
      }
      else {
        if (1 == (int8_t)EEPROM.read(PC))
          bitSource = true;
        PC += 3;                                                                //byte 14
      }

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 17
        tempVarLength = getSource();
      }
      else {
        tempVarLength = (int32_t)getJoinValue();
        PC += 4;                                                                //byte 18
      }

      //Logic
      int tempCalc = 16 - outIndex;
      bool tempPostBit = false;
      bool newBit = bitSource;

      if (tempCalc >= 2) {
        if (tempVarLength > tempCalc) {

          if (ulEnabled)
            ulBit = oState[15];

          for (int j = outIndex; j < 2; j++) {
            tempPostBit = oState[j];
            oState[j] = newBit;
            newBit = tempPostBit;
          }
        }

        else {
          if (ulEnabled)
            ulBit = oState[(outIndex + (tempVarLength - 1))];

          for (int j = outIndex; j < (tempVarLength + outIndex); j++) {
            tempPostBit = oState[j];
            oState[j] = newBit;
            newBit = tempPostBit;
          }
        }
      }//END if(tempCalc >= 2)
      else
        Serial.println("BSL Range Error");

    }//END output

    if (ulEnabled)
      setDnValue(ulBit, ulIndex);
    if (dnEnabled)
      setDnValue(true, dnIndex);

  }//END if(noState)
  else
    PC += 15;
}

void bsr() {                                                                    //Bit Shift Right
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC)) {
    bool tempPre = getUpEdgeActive();
    bool noStateShift = getNoValue();

    if (tempPre)
      noState = true;

    if (!noStateShift) {
      if (isTagEnable(PC + 1))
        setDnValue(false, PC + 1);
    }

  }
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    int dnIndex = 0;
    bool dnEnabled = false;
    int ulIndex = 0;
    bool ulEnabled = false;
    int32_t tempInt = 0;
    int sourceIndex = 0;
    bool bitSource = false;
    int32_t tempVarLength = 0;
    bool ulBit = false;

    PC++;                                                                       //byte 4
    if (isTagEnable(PC)) {                                                      //if dn is enabled
      dnEnabled = true;
      dnIndex = PC;
    }

    PC += 2;                                                                    //byte 6

    if (isTagEnable(PC)) {
      ulEnabled = true;
      ulIndex = PC;
    }

    PC += 2;                                                                    //byte 8

    bool typeVar = isTagEnable(PC);
    PC++;                                                                       //byte 9
    sourceIndex = PC;

    if (typeVar) {                                                              //int-dint
      int32_t tempUl = 0;

      tempInt = (int32_t)getSource();
      PC++;                                                                     //byte 11

      if (isTagEnable(PC + 1)) {
        PC++;                                                                   //byte 12
        bitSource = getNoValue();
        PC++;                                                                   //byte 14
      }
      else {
        if (1 == (int8_t)EEPROM.read(PC))
          bitSource = true;
        PC += 3;                                                                //byte 14
      }

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 17
        tempVarLength = getSource();
      }
      else {
        tempVarLength = (int32_t)getJoinValue();
        PC += 4;                                                                //byte 18
      }



      //Logic
      if (ulEnabled) {
        tempUl = (tempInt & 1);

        if (tempUl == 1)
          ulBit = true;
        else
          ulBit = false;
      }

      int32_t maskUl = 1;
      for (int j = 0; j < (tempVarLength - 1); j++) {
        maskUl = (maskUl << 1);
        maskUl = (maskUl | 1);
      }

      tempUl = tempInt;

      tempUl = tempUl & maskUl;
      tempUl = tempUl >> 1;
      tempUl = tempUl & maskUl;


      if (bitSource)
        tempUl = tempUl | ((int32_t)1 << (tempVarLength - 1));


      int32_t mask2 = ~maskUl;

      tempInt = (tempInt & mask2);

      tempInt = tempInt | tempUl;

      updateVariable(sourceIndex, tempInt);


    }//END int-dint

    else {                                                                      //output
      int outIndex = getOutIndex();
      PC++;                                                                     //byte 11

      if (isTagEnable(PC + 1)) {
        PC++;                                                                   //byte 12
        bitSource = getNoValue();
        PC++;                                                                   //byte 14
      }
      else {
        if (1 == (int8_t)EEPROM.read(PC))
          bitSource = true;
        PC += 3;                                                                //byte 14
      }

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 17
        tempVarLength = getSource();
      }
      else {
        tempVarLength = (int32_t)getJoinValue();
        PC += 4;                                                                //byte 18
      }

      //Logic
      int tempCalc = 16 - outIndex;
      bool tempPostBit = false;
      bool newBit = bitSource;

      if (tempCalc >= 2) {
        if (tempVarLength > tempCalc) {

          if (ulEnabled)
            ulBit = oState[outIndex];

          for (int j = 2; j > outIndex; j--) {
            tempPostBit = oState[j - 1];
            oState[j - 1] = newBit;
            newBit = tempPostBit;
          }
        }

        else {
          if (ulEnabled)
            ulBit = oState[outIndex];

          for (int j = (tempVarLength + outIndex); j > outIndex; j--) {
            tempPostBit = oState[j - 1];
            oState[j - 1] = newBit;
            newBit = tempPostBit;
          }
        }
      }//END if(tempCalc >= 2)
      else
        Serial.println("BSL Range Error");

    }//END output

    if (ulEnabled)
      setDnValue(ulBit, ulIndex);
    if (dnEnabled)
      setDnValue(true, dnIndex);

  }//END if(noState)
  else
    PC += 15;
}

void rol() {                                                                    //Rotate left
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC)) {
    bool tempPre = getUpEdgeActive();
    bool noStateShift = getNoValue();

    if (tempPre)
      noState = true;

    if (!noStateShift) {
      if (isTagEnable(PC + 1))
        setDnValue(false, PC + 1);
    }

  }
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    int dnIndex = 0;
    bool dnEnabled = false;
    int ulIndex = 0;
    bool ulEnabled = false;
    int32_t tempInt = 0;
    int sourceIndex = 0;
    int32_t rotate = 0;
    int32_t tempVarLength = 0;
    bool ulBit = false;

    PC++;                                                                       //byte 4
    if (isTagEnable(PC)) {                                                      //if dn is enabled
      dnEnabled = true;
      dnIndex = PC;
    }

    PC += 2;                                                                    //byte 6

    if (isTagEnable(PC)) {
      ulEnabled = true;
      ulIndex = PC;
    }

    PC += 2;                                                                    //byte 8

    bool typeVar = isTagEnable(PC);
    PC++;                                                                       //byte 9
    sourceIndex = PC;

    if (typeVar) {                                                              //int-dint

      int32_t tempUl = 0;
      int32_t tempU2 = 0;

      tempInt = (int32_t)getSource();
      PC++;                                                                     //byte 11

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 14
        rotate = (int32_t)getSource();
        PC++;                                                                   //byte 16
      }
      else {
        rotate = (int32_t)getJoinValue();
        PC += 5;                                                                //byte 16
      }

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 19
        tempVarLength = getSource();
      }
      else {
        tempVarLength = (int32_t)getJoinValue();
        PC += 4;                                                                //byte 20
      }

      //Logic

      int32_t maskUl = 1;
      for (int j = 0; j < (tempVarLength - 1); j++) {
        maskUl = (maskUl << 1);
        maskUl = (maskUl | 1);
      }

      tempUl = tempInt;

      tempUl = tempUl & maskUl;

      int32_t tempPreBit = 0;

      for (int k = 0; k < rotate; k++) {
        tempU2 = (tempUl & ((int32_t)1 << (tempVarLength - 1)));
        tempPreBit = (tempU2 >> (tempVarLength - 1));

        tempUl = tempUl << 1;
        tempUl = tempUl & maskUl;
        tempUl = tempUl | tempPreBit;
      }

      if (ulEnabled) {
        tempU2 = (tempUl & ((int32_t)1 << (tempVarLength - 1)));

        tempU2 = (tempU2 >> (tempVarLength - 1));
        if (tempU2 == 1)
          ulBit = true;
        else
          ulBit = false;
      }

      maskUl = ~maskUl;

      tempInt = (tempInt & maskUl);

      tempInt = tempInt | tempUl;

      updateVariable(sourceIndex, tempInt);


    }//END int-dint

    else {                                                                      //output
      int outIndex = getOutIndex();
      PC++;                                                                     //byte 11

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 14
        rotate = (int32_t)getSource();
        PC++;                                                                   //byte 16
      }
      else {
        rotate = (int32_t)getJoinValue();
        PC += 5;                                                                //byte 16
      }

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 19
        tempVarLength = getSource();
      }
      else {
        tempVarLength = (int32_t)getJoinValue();
        PC += 4;                                                                //byte 20
      }

      //Logic
      int tempCalc = 16 - outIndex;
      bool tempPostBit = false;

      if (tempCalc >= 2) {
        if (tempVarLength > tempCalc) {
          bool newBit = false;

          for (int k = 0; k < rotate; k++) {
            for (int j = outIndex; j < 2; j++) {
              tempPostBit = oState[j];
              oState[j] = newBit;
              newBit = tempPostBit;
            }
            oState[outIndex] = newBit;
          }

          if (ulEnabled)
            ulBit = oState[15];

        }

        else {

          bool newBit = false;

          for (int k = 0; k < rotate; k++) {
            for (int j = outIndex; j < (tempVarLength + outIndex); j++) {
              tempPostBit = oState[j];
              oState[j] = newBit;
              newBit = tempPostBit;
            }
            oState[outIndex] = newBit;
          }

          if (ulEnabled)
            ulBit = oState[(tempVarLength + (outIndex - 1))];

        }
      }//END if(tempCalc >= 2)
      else
        Serial.println("ROL Range Error");

    }//END output

    if (ulEnabled)
      setDnValue(ulBit, ulIndex);
    if (dnEnabled)
      setDnValue(true, dnIndex);

  }//END if(noState)
  else
    PC += 17;
}

void ror() {                                                                    //Rotate Right
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC)) {
    bool tempPre = getUpEdgeActive();
    bool noStateShift = getNoValue();

    if (tempPre)
      noState = true;

    if (!noStateShift) {
      if (isTagEnable(PC + 1))
        setDnValue(false, PC + 1);
    }

  }
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    int dnIndex = 0;
    bool dnEnabled = false;
    int ulIndex = 0;
    bool ulEnabled = false;
    int32_t tempInt = 0;
    int sourceIndex = 0;
    int32_t rotate = 0;
    int32_t tempVarLength = 0;
    bool ulBit = false;

    PC++;                                                                       //byte 4
    if (isTagEnable(PC)) {                                                      //if dn is enabled
      dnEnabled = true;
      dnIndex = PC;
    }

    PC += 2;                                                                    //byte 6

    if (isTagEnable(PC)) {
      ulEnabled = true;
      ulIndex = PC;
    }

    PC += 2;                                                                    //byte 8

    bool typeVar = isTagEnable(PC);
    PC++;                                                                       //byte 9
    sourceIndex = PC;

    if (typeVar) {                                                              //int-dint

      int32_t tempUl = 0;
      int32_t tempU2 = 0;

      tempInt = (int32_t)getSource();
      PC++;                                                                     //byte 11

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 14
        rotate = (int32_t)getSource();
        PC++;                                                                   //byte 16
      }
      else {
        rotate = (int32_t)getJoinValue();
        PC += 5;                                                                //byte 16
      }

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 19
        tempVarLength = getSource();
      }
      else {
        tempVarLength = (int32_t)getJoinValue();
        PC += 4;                                                                //byte 20
      }

      //Logic

      int32_t maskUl = 1;
      for (int j = 0; j < (tempVarLength - 1); j++) {
        maskUl = (maskUl << 1);
        maskUl = (maskUl | 1);
      }

      tempUl = tempInt;

      tempUl = tempUl & maskUl;

      int32_t tempPreBit = 0;

      for (int k = 0; k < rotate; k++) {
        tempPreBit =  (tempUl & 1);
        tempPreBit =  (tempPreBit << (tempVarLength - 1));

        tempUl = tempUl >> 1;
        tempUl = tempUl & maskUl;
        tempUl = tempUl | tempPreBit;
      }

      if (ulEnabled) {
        tempU2 = (tempPreBit >> (tempVarLength - 1));
        if (tempU2 == 1)
          ulBit = true;
        else
          ulBit = false;
      }

      maskUl = ~maskUl;

      tempInt = (tempInt & maskUl);

      tempInt = tempInt | tempUl;

      updateVariable(sourceIndex, tempInt);


    }//END int-dint

    else {                                                                      //output
      int outIndex = getOutIndex();
      PC++;                                                                     //byte 11

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 14
        rotate = (int32_t)getSource();
        PC++;                                                                   //byte 16
      }
      else {
        rotate = (int32_t)getJoinValue();
        PC += 5;                                                                //byte 16
      }

      if (isTagEnable(PC + 3)) {
        PC += 3;                                                                //byte 19
        tempVarLength = getSource();
      }
      else {
        tempVarLength = (int32_t)getJoinValue();
        PC += 4;                                                                //byte 20
      }

      //Logic
      int tempCalc = 16 - outIndex;
      bool tempPostBit = false;

      if (tempCalc >= 2) {
        if (tempVarLength > tempCalc) {
          bool newBit = false;

          for (int k = 0; k < rotate; k++) {
            for (int j = 2; j > outIndex; j--) {
              tempPostBit = oState[j - 1];
              oState[j - 1] = newBit;
              newBit = tempPostBit;
            }
            oState[15] = newBit;
          }

          if (ulEnabled)
            ulBit = oState[15];

        }

        else {
          bool newBit = false;

          for (int k = 0; k < rotate; k++) {
            for (int j = (tempVarLength + outIndex); j > outIndex; j--) {
              tempPostBit = oState[j - 1];
              oState[j - 1] = newBit;
              newBit = tempPostBit;
            }
            oState[tempVarLength + (outIndex - 1)] = newBit;
          }

          if (ulEnabled)
            ulBit = oState[(tempVarLength + (outIndex - 1))];

        }
      }//END if(tempCalc >= 2)
      else
        Serial.println("ROR Range Error");

    }//END output

    if (ulEnabled)
      setDnValue(ulBit, ulIndex);
    if (dnEnabled)
      setDnValue(true, dnIndex);

  }//END if(noState)
  else
    PC += 17;
}

void scl() {                                                                    //Scale
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    int outputIndex = 0;
    float tempInput = 0;
    float tempRate = 0;
    float tempOffset = 0;


    PC++;                                                                     //byte 4
    outputIndex = PC;
    PC += 2;                                                                  //byte 6

    tempInput = getSource();
    PC++;                                                                     //byte 8

    if (isTagEnable(PC + 3)) {
      PC += 3;                                                                //byte 11
      tempRate = getSource();
      PC++;                                                                   //byte 13
    }
    else {
      tempRate = (float)getJoinValue();
      PC += 5;                                                                //byte 13
    }

    if (isTagEnable(PC + 3)) {
      PC += 3;                                                                //byte 16
      tempOffset = getSource();
    }
    else {
      tempOffset = (float)getJoinValue();
      PC += 4;                                                                //byte 17
    }

    float tempResult = tempRate / 10000L;
    tempResult *= tempInput;
    tempResult += tempOffset;


    setDest(outputIndex, tempResult);


  }
  else
    PC += 14;
}

void scp() {                                                                    //Scale with parameter
  bool noState = false;

  PC++;                                                                         //byte 2

  if (isTagEnable(PC))
    noState = getNoValue();
  else {
    noState = true;
    PC++;                                                                       //byte 3
  }


  if (noState) {
    int outputIndex = 0;
    float tempInput = 0;
    float tempInputMin = 0;
    float tempInputMax = 0;
    float tempOutputMin = 0;
    float tempOutputMax = 0;

    PC++;                                                                     //byte 4
    outputIndex = PC;
    PC += 2;                                                                  //byte 6

    tempInput = getSource();
    PC++;                                                                     //byte 8

    if (isTagEnable(PC + 3)) {
      PC += 3;                                                                //byte 11
      tempInputMin = getSource();
      PC++;                                                                   //byte 13
    }
    else {
      tempInputMin = (float)getJoinValue();
      PC += 5;                                                                //byte 13
    }

    if (isTagEnable(PC + 3)) {
      PC += 3;                                                                //byte 16
      tempInputMax = getSource();
      PC++;                                                                   //byte 18
    }
    else {
      tempInputMax = (float)getJoinValue();
      PC += 5;                                                                //byte 18
    }


    if (isTagEnable(PC + 3)) {
      PC += 3;                                                                //byte 21
      tempOutputMin = getSource();
      PC++;                                                                   //byte 23
    }
    else {
      tempOutputMin = (float)getJoinValue();
      PC += 5;                                                                //byte 23
    }

    if (isTagEnable(PC + 3)) {
      PC += 3;                                                                //byte 26
      tempOutputMax = getSource();
    }
    else {
      tempOutputMax = (float)getJoinValue();
      PC += 4;                                                                //byte 27
    }

    float tempResult = 0;
    bool calculatePlease = true;

    if (tempInputMax == tempInputMin)
      calculatePlease = false;
    if (tempOutputMin == tempOutputMax)
      calculatePlease = false;
    if (tempInputMin > tempInputMax)
      calculatePlease = false;
    if (tempOutputMin > tempOutputMax)
      calculatePlease = false;

    if (calculatePlease) {
      float tempRateCalc = ((tempOutputMax - tempOutputMin) / (tempInputMax - tempInputMin));
      float tempOffsetCalc = tempOutputMin - (tempInputMin * tempRateCalc);

      tempResult = (tempInput * tempRateCalc) + tempOffsetCalc;
    }
    else
      tempResult = tempOutputMin;

    if (tempResult > tempOutputMax)
      tempResult = tempOutputMax;
    else if (tempResult < tempOutputMin)
      tempResult = tempOutputMin;

    setDest(outputIndex, tempResult);
    //Serial.println("SCP");


  }
  else
    PC += 24;
}

bool isTagEnable(int tempPC) {
  bool tempBool = false;

  if (EEPROM.read(tempPC) != 0)
    tempBool = true;

  return tempBool;
}

int16_t getJoinValue() {
  int16_t tempA = 0;
  int16_t tempB = 0;

  tempB = (int16_t)EEPROM.read(PC) << 14;
  tempA = (int16_t)EEPROM.read(PC + 1) << 7;
  tempB = tempB | tempA | (int16_t)EEPROM.read(PC + 2);

  return tempB;
}

int getOutIndex() {
  int tempReturn = 0;

  uint8_t myDataByte = EEPROM.read(PC);
  PC++;
  uint8_t n = EEPROM.read(PC);
  n = n - 48;

  switch (myDataByte) {
    case 'O':                                                                   //output
      tempReturn = n;
      break;
    case 'k':                                                                   //output + 10
      tempReturn = n + 10;
      break;
  }

  return tempReturn;
}

void setDnValue(bool newState, int tempPC) {

  uint8_t myDataByte = EEPROM.read(tempPC);
  uint8_t n = EEPROM.read(tempPC + 1);
  n = n - 48;

  switch (myDataByte) {
    case 'O':                                                                   //output
      oState[n] = newState;
      break;
    case 'k':                                                                   //output + 10
      oState[n + 10] = newState;
      break;
    case 'V':                                                                   //virtual
      vState[n] = newState;
      break;
    case 'v':                                                                   //virtual + 10
      vState[n + 10] = newState;
      break;
    case 'U':                                                                   //virtual + 20
      vState[n + 20] = newState;
      break;
    case 'u':                                                                   //virtual + 30
      vState[n + 30] = newState;
      break;
    case 'A':                                                                   //virtual + 40
      vState[n + 40] = newState;
      break;
    case 'a':                                                                   //virtual + 50
      vState[n + 50] = newState;
      break;
    case 'c':                                                                   //virtual + 60
      vState[n + 60] = newState;
      break;
  }

}

bool getUpEdgeActive() {
  bool tempReturn = false;

  uint8_t myDataByte = EEPROM.read(PC);
  uint8_t n = EEPROM.read(PC + 1);
  n = n - 48;

  if ((myDataByte) != 0) {
    switch (myDataByte) {
      case 'I':                                                                 //intput
        tempReturn = iUpEdge[n];
        break;
      case 'i':                                                                 //input + 10
        tempReturn = iUpEdge[n + 10];
        break;
      case 'O':                                                                   //output
        tempReturn = oUpEdge[n];
        break;
      case 'k':                                                                   //output + 10
        tempReturn = oUpEdge[n + 10];
        break;
      case 'V':                                                                   //virtual
        tempReturn = vUpEdge[n];
        break;
      case 'v':                                                                   //virtual + 10
        tempReturn = vUpEdge[n + 10];
        break;
      case 'U':                                                                   //virtual + 20
        tempReturn = vUpEdge[n + 20];
        break;
      case 'u':                                                                   //virtual + 30
        tempReturn = vUpEdge[n + 30];
        break;
      case 'A':                                                                   //virtual + 40
        tempReturn = vUpEdge[n + 40];
        break;
      case 'a':                                                                   //virtual + 50
        tempReturn = vUpEdge[n + 50];
        break;
      case 'c':                                                                   //virtual + 60
        tempReturn = vUpEdge[n + 60];
        break;

    }
  }
  else
    tempReturn = true;

  return tempReturn;
}

bool getNoValue() {
  bool tempReturn = false;

  uint8_t myDataByte = EEPROM.read(PC);
  PC++;
  uint8_t n = EEPROM.read(PC);
  n = n - 48;


  if ((myDataByte) != 0) {
    switch (myDataByte) {
      case 'I':                                                                 //intput
        tempReturn = iState[n];
        break;
      case 'i':                                                                 //input + 10
        tempReturn = iState[n + 10];
        break;
      case 'O':                                                                   //output
        tempReturn = oState[n];
        break;
      case 'k':                                                                   //output + 10
        tempReturn = oState[n + 10];
        break;
      case 'V':                                                                   //virtual
        tempReturn = vState[n];
        break;
      case 'v':                                                                   //virtual + 10
        tempReturn = vState[n + 10];
        break;
      case 'U':                                                                   //virtual + 20
        tempReturn = vState[n + 20];
        break;
      case 'u':                                                                   //virtual + 30
        tempReturn = vState[n + 30];
        break;
      case 'A':                                                                   //virtual + 40
        tempReturn = vState[n + 40];
        break;
      case 'a':                                                                   //virtual + 50
        tempReturn = vState[n + 50];
        break;
      case 'c':                                                                   //virtual + 60
        tempReturn = vState[n + 60];
        break;

    }
  }
  else
    tempReturn = true;

  return tempReturn;
}

bool isSourceInt() {
  bool tempReturn = false;

  uint8_t myDataByte = EEPROM.read(PC);

  switch (myDataByte) {
    case 'D':                                                                   //int
      tempReturn = true;
      break;
    case 'd':                                                                   //int + 10
      tempReturn = true;
      break;
    case 'M':                                                                   //int + 20
      tempReturn = true;
      break;
    case 'N':                                                                   //int + 30
      tempReturn = true;
      break;
    case 'y':                                                                   //int time                                    //NEW_UPDATE
      tempReturn = true;
      break;
  }

  return tempReturn;
}

float getSource() {
  float tempReturn = 0;

  uint8_t myDataByte = EEPROM.read(PC);
  PC++;
  uint8_t n = EEPROM.read(PC);
  n = n - 48;

  switch (myDataByte) {
    case 'p':                                                                   //analog input
      tempReturn = (float)readAnalogPin(n);
      break;
    case 'Q':                                                                   //analog input + 10
      tempReturn = (float)readAnalogPin(n + 10);
      break;
    case 'o':                                                                   //analog output
      tempReturn = (float)readAnalogOutput(n);
      break;
    case 'q':                                                                   //analog output + 10
      tempReturn = (float)readAnalogOutput(n + 10);
      break;
    case 'D':                                                                   //int
      tempReturn = (float)intValues[n];
      break;
    case 'd':                                                                   //int + 10
      tempReturn = (float)intValues[n + 10];
      break;
    case 'M':                                                                   //int + 20
      tempReturn = (float)intValues[n + 20];
      break;
    case 'N':                                                                   //int + 30
      tempReturn = (float)intValues[n + 30];
      break;
    case 'y':
      tempReturn = (float)timeData[n];                                          //int time                                        //NEW_UPDATE
      break;
    case 'L':                                                                   //long
      tempReturn = (float)dintValues[n];
      break;
    case 'H':                                                                   //long + 10
      tempReturn = (float)dintValues[n + 10];
      break;
    case 'h':                                                                   //long + 20
      tempReturn = (float)dintValues[n + 20];
      break;
    case 'J':                                                                   //long + 30
      tempReturn = (float)dintValues[n + 30];
      break;
    case 'j':                                                                   //long + 40
      tempReturn = (float)dintValues[n + 40];
      break;
    case 'K':                                                                   //long + 50
      tempReturn = (float)dintValues[n + 50];
      break;
    case 'F':                                                                   //real
      tempReturn = realValues[n];
      break;
    case 'E':                                                                   //real + 10
      tempReturn = realValues[n + 10];
      break;
    case 'g':                                                                   //real + 20
      tempReturn = realValues[n + 20];
      break;
    case 'G':                                                                   //real + 30
      tempReturn = realValues[n + 30];
      break;
    default:
      tempReturn = 0;
      break;
  }

  return tempReturn;
}

void updateVariable(int tempPC, int32_t valueDest) {
  uint8_t myDataByte = EEPROM.read(tempPC);
  uint8_t n = EEPROM.read(tempPC + 1);
  n = n - 48;

  switch (myDataByte) {
    case 'D':                                                                   //int
      intValues[n] = (int16_t)valueDest;
      break;
    case 'd':                                                                   //int + 10
      intValues[n + 10] = (int16_t)valueDest;
      break;
    case 'M':                                                                   //int + 20
      intValues[n + 20] = (int16_t)valueDest;
      break;
    case 'N':                                                                   //int + 30
      intValues[n + 30] = (int16_t)valueDest;
      break;
    case 'L':                                                                   //long
      dintValues[n] = valueDest;
      break;
    case 'H':                                                                   //long + 10
      dintValues[n + 10] = valueDest;
      break;
    case 'h':                                                                   //long + 20
      dintValues[n + 20] = valueDest;
      break;
    case 'J':                                                                   //long + 30
      dintValues[n + 30] = valueDest;
      break;
    case 'j':                                                                   //long + 40
      dintValues[n + 40] = valueDest;
      break;
    case 'K':                                                                   //long + 50
      dintValues[n + 50] = valueDest;
      break;

  }
}

void setDest(float valueDest) {

  uint8_t myDataByte = EEPROM.read(PC);
  PC++;
  uint8_t n = EEPROM.read(PC);
  n = n - 48;

  switch (myDataByte) {
    case 'x':
      writeServoPin(n, (int)valueDest);
      break;
    case 'X':
      writeServoPin(n + 10, (int)valueDest);
      break;
    case 'o':                                                                   //Analog output
      writeAnalogPin(n, (int)valueDest);
      break;
    case 'q':                                                                   //Analog output + 10
      writeAnalogPin(n + 10, (int)valueDest);
      break;
    case 'D':                                                                   //int
      intValues[n] = (int16_t)valueDest;
      break;
    case 'd':                                                                   //int + 10
      intValues[n + 10] = (int16_t)valueDest;
      break;
    case 'M':                                                                   //int + 20
      intValues[n + 20] = (int16_t)valueDest;
      break;
    case 'N':                                                                   //int + 30
      intValues[n + 30] = (int16_t)valueDest;
      break;
    case 'L':                                                                   //long
      dintValues[n] = (int32_t)valueDest;
      break;
    case 'H':                                                                   //long + 10
      dintValues[n + 10] = (int32_t)valueDest;
      break;
    case 'h':                                                                   //long + 20
      dintValues[n + 20] = (int32_t)valueDest;
      break;
    case 'J':                                                                   //long + 30
      dintValues[n + 30] = (int32_t)valueDest;
      break;
    case 'j':                                                                   //long + 40
      dintValues[n + 40] = (int32_t)valueDest;
      break;
    case 'K':                                                                   //long + 50
      dintValues[n + 50] = (int32_t)valueDest;
      break;
    case 'F':                                                                   //real
      realValues[n] = valueDest;
      break;
    case 'E':                                                                   //real + 10
      realValues[n + 10] = valueDest;
      break;
    case 'g':                                                                   //real + 20
      realValues[n + 20] = valueDest;
      break;
    case 'G':                                                                   //real + 30
      realValues[n + 30] = valueDest;
      break;

  }

}

void setDest(int tempPC, float valueDest) {

  uint8_t myDataByte = EEPROM.read(tempPC);
  uint8_t n = EEPROM.read(tempPC + 1);
  n = n - 48;

  switch (myDataByte) {
    case 'x':
      writeServoPin(n, (int)valueDest);
      break;
    case 'X':
      writeServoPin(n + 10, (int)valueDest);
      break;
    case 'o':                                                                   //Analog output
      writeAnalogPin(n, (int)valueDest);
      break;
    case 'q':                                                                   //Analog output + 10
      writeAnalogPin(n + 10, (int)valueDest);
      break;
    case 'D':                                                                   //int
      intValues[n] = (int16_t)valueDest;
      break;
    case 'd':                                                                   //int + 10
      intValues[n + 10] = (int16_t)valueDest;
      break;
    case 'M':                                                                   //int + 20
      intValues[n + 20] = (int16_t)valueDest;
      break;
    case 'N':                                                                   //int + 30
      intValues[n + 30] = (int16_t)valueDest;
      break;
    case 'L':                                                                   //long
      dintValues[n] = (int32_t)valueDest;
      break;
    case 'H':                                                                   //long + 10
      dintValues[n + 10] = (int32_t)valueDest;
      break;
    case 'h':                                                                   //long + 20
      dintValues[n + 20] = (int32_t)valueDest;
      break;
    case 'J':                                                                   //long + 30
      dintValues[n + 30] = (int32_t)valueDest;
      break;
    case 'j':                                                                   //long + 40
      dintValues[n + 40] = (int32_t)valueDest;
      break;
    case 'K':                                                                   //long + 50
      dintValues[n + 50] = (int32_t)valueDest;
      break;
    case 'F':                                                                   //real
      realValues[n] = valueDest;
      break;
    case 'E':                                                                   //real + 10
      realValues[n + 10] = valueDest;
      break;
    case 'g':                                                                   //real + 20
      realValues[n + 20] = valueDest;
      break;
    case 'G':                                                                   //real + 30
      realValues[n + 30] = valueDest;
      break;

  }

}

int readMyByteUntil(int delimi, WiFiClient &serTemp1) {
  inCount = true;
  int countMeBuff = 0;
  int myCountBuff = 0;
  waitReadCounter = 0;

  while (inCount) {
    int inSize = serTemp1.available();
    if (inSize > 0) {

      for (int i = 0; i < inSize; i++) {

        int tempBuffData = serTemp1.read();

        if (tempBuffData != delimi) {
          bufferSerial[i + waitReadCounter] = tempBuffData;
          countMeBuff++;
        }

        else {
          inCount = false;

        }

      }
      waitReadCounter += countMeBuff;
    }


    if (myCountBuff > 19) {
      inCount = false;
    }

    myCountBuff++;
    delay(50);
  }//END WHILE

  return waitReadCounter;
}

void fragmentData(uint32_t tempInt32, uint8_t basura) {
  ld = (tempInt32 & (BYTE6MASK0));

  tempValue32 = (tempInt32 & (BYTE6MASK1));
  md = (tempValue32 >> 6);

  tempValue32 = (tempInt32 & (BYTE6MASK2));
  hd = (tempValue32 >> 12);

  tempValue32 = (tempInt32 & (BYTE6MASK3));
  ld32 = (tempValue32 >> 18);

  tempValue32 = (tempInt32 & (BYTE6MASK4));
  md32 = (tempValue32 >> 24);

  tempValue32 = (tempInt32 & (BYTE6MASK5));
  hd32 = (tempValue32 >> 30);

}

void fragmentData(uint32_t tempInt16) {
  ld = (tempInt16 & (BYTE6MASK0));

  tempValue = (tempInt16 & (MDMASK));
  md = (tempValue >> 6);

  tempValue = (tempInt16 & (HDMASK));
  hd = (tempValue >> 12);
}


void fragmentData(uint32_t tempInt16, boolean bol, boolean bol2) {
  ld = (tempInt16 & (BYTE6MASK0));

  tempValue = (tempInt16 & (MDMASK));
  md = (tempValue >> 6);
}

uint32_t joinInt32(char temp6, char temp5, char temp4, char temp3, char temp2, char temp1) {
  uint32_t tempA = 0;
  uint32_t tempB = 0;

  tempB = (((int32_t)temp6) << 30);
  tempA = (((int32_t)temp5) << 24);
  tempB = tempB | tempA;
  tempA = (((int32_t)temp4) << 18);
  tempB = tempB | tempA;
  tempA = (((int32_t)temp3) << 12);
  tempB = tempB | tempA;
  tempA = (((int32_t)temp2) << 6);
  tempB = tempB | tempA | ((int32_t)temp1);

  return tempB;
}

float joinFloat(uint32_t tempInt32) {
  float temp = 0;
  floatValue tempFloat;

  tempFloat.bytes[0] = (uint8_t)(tempInt32 & BYTEMASK0);
  tempFloat.bytes[1] = (uint8_t)((tempInt32 & BYTEMASK1) >> 8);
  tempFloat.bytes[2] = (uint8_t)((tempInt32 & BYTEMASK2) >> 16);
  tempFloat.bytes[3] = (uint8_t)((tempInt32 & BYTEMASK3) >> 24);

  temp = tempFloat.val;

  return temp;
}


uint16_t joinInt16(char tempHd, char tempMd, char tempLd) {
  uint16_t temp = 0;

  uint16_t tempLd1 = (int)tempLd;
  uint16_t tempMd1 = (int)tempMd;
  uint16_t tempHd1 = (int)tempHd;

  tempMd1 = (tempMd1 << 6);
  tempHd1 = (tempHd1 << 12);

  temp = (tempLd1 | tempMd1 | tempHd1);
  temp = temp & GMASK16;

  return temp;
}

uint8_t joinInt8(char tempMd, char tempLd) {
  uint8_t temp = 0;

  uint8_t tempLd1 = (int)tempLd;
  uint8_t tempMd1 = (int)tempMd;

  tempMd1 = (tempMd1 << 6);

  temp = (tempLd1 | tempMd1);
  temp = temp & GMASK8;

  return temp;
}

void cleanClientData(WiFiServer &serTemp1) {
  WiFiClient serTemp2 = serTemp1.available();

  if (serTemp2.available()) {
    while (serTemp2.read() != -1) {

    }
  }
}

void checkSpace() {
  PC = 0; //put the PC in the start of the memory
  int dataByte = EEPROM.read(PC);
  Serial.println("");
  Serial.write(dataByte);
  while (dataByte != 'z') { //z is the value that indicate that the program has ended
    PC++;
    dataByte = EEPROM.read(PC);
    if ((dataByte == 'R') || (dataByte == '$'))
      Serial.println("");
    Serial.write(dataByte);
  }

  dataByte = PC + 1;
  int tempPC = (100 * (dataByte)) / 3285;


  Serial.println("");
  Serial.print("Program bytes: ");
  Serial.print(PC + 1);
  Serial.print(" (");
  Serial.print(tempPC);
  Serial.println("%)");


  Serial.print("Free: ");
  tempPC = 100 - tempPC;
  Serial.print(tempPC);
  Serial.println("%");
  Serial.println(" ");
}

// send an NTP request to the time server at the given address
void sendNTPpacket(char* address) {
  // set all bytes in the buffer to 0
  memset(packetBuffer, 0, NTP_PACKET_SIZE);
  // Initialize values needed to form NTP request
  // (see URL above for details on the packets)
  packetBuffer[0] = 0b11100011;   // LI, Version, Mode
  packetBuffer[1] = 0;     // Stratum, or type of clock
  packetBuffer[2] = 6;     // Polling Interval
  packetBuffer[3] = 0xEC;  // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12]  = 49;
  packetBuffer[13]  = 0x4E;
  packetBuffer[14]  = 49;
  packetBuffer[15]  = 52;

  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp:
  Udp.beginPacket(address, 123); //NTP requests are to port 123
  Udp.write(packetBuffer, NTP_PACKET_SIZE);
  Udp.endPacket();
}


int setGMT(int timeOffset) {
  int secondsOffset = 0;
  boolean negativeValue = false;

  if (timeOffset < 0) {
    negativeValue = true;
    timeOffset *= (-1);
  }

  switch (timeOffset) {
    case 100:         //GMT+1:00
      secondsOffset = 3600;
      break;
    case 200:         //GMT+2:00
      secondsOffset = 7200;
      break;
    case 230:         //GMT+2:30
      secondsOffset = 9000;
      break;
    case 300:         //GMT+3:00
      secondsOffset = 10800;
      break;
    case 400:         //GMT+4:00
      secondsOffset = 14400;
      break;
    case 430:         //GMT+4:30
      secondsOffset = 16200;
      break;
    case 500:         //GMT+5:00
      secondsOffset = 18000;
      break;
    case 530:         //GMT+5:30
      secondsOffset = 19800;
      break;
    case 545:         //GMT+5:45
      secondsOffset = 20700;
      break;
    case 600:         //GMT+6:00
      secondsOffset = 21600;
      break;
    case 630:         //GMT+6:30
      secondsOffset = 23400;
      break;
    case 700:         //GMT+7:00
      secondsOffset = 25200;
      break;
    case 800:         //GMT+8:00
      secondsOffset = 28800;
      break;
    case 830:         //GMT+8:30
      secondsOffset = 30600;
      break;
    case 845:         //GMT+8:45
      secondsOffset = 31500;
      break;
    case 900:         //GMT+9:00
      secondsOffset = 32400;
      break;
    case 930:         //GMT+9:30
      secondsOffset = 34200;
      break;
    case 1000:        //GMT+10:00
      secondsOffset = 36000;
      break;
    case 1030:        //GMT+10:30
      secondsOffset = 37800;
      break;
    case 1100:        //GMT+11:00
      secondsOffset = 39600;
      break;
    case 1200:        //GMT+12:00
      secondsOffset = 43200;
      break;
    case 1245:        //GMT+12:45
      secondsOffset = 45900;
      break;
    case 1300:        //GMT+13:00
      secondsOffset = 46800;
      break;
    case 1400:        //GMT+14:00
      secondsOffset = 50400;
      break;
  }

  if (negativeValue)
    secondsOffset *= (-1);


  return secondsOffset;
}



void updateTimeRegisters() {
  timeData[0] = year();
  timeData[1] = month();
  timeData[2] = weekday();
  timeData[3] = day();
  timeData[4] = hour();
  timeData[5] = minute();
  timeData[6] = second();
}


void updateAllTime() {
  setTime(timeData[4], timeData[5], timeData[6], timeData[3], timeData[1] + 1, timeData[0]);

}


void commitToEEPROM() {
  for (int i = 0; i < 4096; i++) {
    EEPROM.write(i, preEEPROM[i]);
  }

  EEPROM.commit();
}

void startWifi() {

  Serial.begin(115200);

  WiFi.begin(ssid, password);

  bool connectingWiFi = true;
  int countAttempts = 0;

  M5.Lcd.print(" Connecting");

  while (connectingWiFi) {
    if (WiFi.status() == WL_CONNECTED) {
      connectingWiFi = false;
      wifiConnected = true;
    }

    if (countAttempts >= 20)
      connectingWiFi = false;

    delay(500);
    countAttempts++;
    M5.Lcd.print(".");
  }

  if (wifiConnected) {

    M5.Lcd.println();
    M5.Lcd.println();
    M5.Lcd.print(" Connected :) - IP address: ");
    M5.Lcd.print(WiFi.localIP());
    M5.Lcd.println();


    getTime();      

    server.begin();
    server.setNoDelay(false);
  }
  else{
    M5.Lcd.println();
    M5.Lcd.println();
    M5.Lcd.print(" Not connected :(");
    M5.Lcd.println();
  }

  delay(800);

  M5.Lcd.println();
  M5.Lcd.print(" RTC: ");
  M5.Lcd.print(year());
  M5.Lcd.print('/');
  M5.Lcd.print(month());
  M5.Lcd.print('/');
  M5.Lcd.print(day());
  M5.Lcd.print("  ");
  M5.Lcd.print(hour());
  M5.Lcd.print(':');
  M5.Lcd.print(minute());
  M5.Lcd.print(':');
  M5.Lcd.print(second());
  M5.Lcd.print("  ");
  M5.Lcd.println(weekday());
  M5.Lcd.println();

  delay(2000);
}

void getTime() {

  Udp.begin(udpLocalPort);

  sendNTPpacket(timeServer); // send an NTP packet to a time server

  delay(1000);

  if (Udp.parsePacket()) {
    // We've received a packet, read the data from it
    Udp.read(packetBuffer, NTP_PACKET_SIZE); // read the packet into the buffer

    // the timestamp starts at byte 40 of the received packet and is four bytes,
    // or two words, long. First, extract the two words:

    unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
    unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
    // combine the four bytes (two words) into a long integer
    // this is NTP time (seconds since Jan 1 1900):
    unsigned long secsSince1900 = highWord << 16 | lowWord;



    const unsigned long seventyYears = 2208988800UL;
    epoch = secsSince1900 - seventyYears;
    epoch += setGMT(clockGMT);

    setTime(epoch);

    rtcUpdated = true;

  }

}

void initADC(){
  analogReadResolution(12);
  analogSetClockDiv(1);
  analogSetAttenuation(ADC_11db);
}

void initAnalogOut(){
  ledcSetup(0, LEDC_BASE_FREQ_PWM, 12);
  ledcSetup(1, LEDC_BASE_FREQ_PWM, 12);
  ledcSetup(2, LEDC_BASE_FREQ_PWM, 12);
  ledcSetup(3, LEDC_BASE_FREQ_PWM, 12);
  ledcSetup(4, LEDC_BASE_FREQ_PWM, 12);
  ledcSetup(5, LEDC_BASE_FREQ_PWM, 12);
  ledcSetup(6, LEDC_BASE_FREQ_PWM, 12);
  ledcSetup(7, LEDC_BASE_FREQ_PWM, 12);

  ledcSetup(8, LEDC_BASE_SERVO, 10);
  ledcSetup(9, LEDC_BASE_SERVO, 10);
  ledcSetup(10, LEDC_BASE_SERVO, 10);
  ledcSetup(11, LEDC_BASE_SERVO, 10);
  ledcSetup(12, LEDC_BASE_SERVO, 10);
  ledcSetup(13, LEDC_BASE_SERVO, 10);
  ledcSetup(14, LEDC_BASE_SERVO, 10);
  ledcSetup(15, LEDC_BASE_SERVO, 10);
  
}


void writeServo(uint8_t channel, int angle){
  float pwmDuty = ((angle*97)/(180))+26;

  ledcWrite(channel, pwmDuty); 
}

void addAnalogInputIndex(int index){
  analogIEnabled[index] = true;
}

void updateUI(){

  int tempValue;
  
  for(int i=0; i<3; i++){
    if(analogIEnabled[i]){
      if(analogInputs[i] != preAnalogInputs[i]){
        preAnalogInputs[i] = analogInputs[i]; 
        analogPinI(analogInputs[i], i);  
      }      
    }
    else{

      if(iState[i] != preIState[i]){
        preIState[i] = iState[i];
        digitalPinI(iState[i], i);
      }      
    }
  }//END for


  for(int i=0; i<4; i++){
    if((pwmOutput[i]) || (servoOutput[i])){
      bool isServo = true;
      if(pwmOutput[i])
        isServo = false;

      if(analogOutputs[i] != preAnalogOutputs[i]){
        preAnalogOutputs[i] = analogOutputs[i];
        analogPinO(analogOutputs[i], i, isServo);
      }
  
    }
    else{
      if(oState[i] != preOState[i]){
        preOState[i] = oState[i];
        digitalPinO(oState[i],i);
      }
    }
  }//END for
  
}


void checkInputBtn(){

  if(M5.BtnA.pressedFor(2000)){
    inReprogramTime = false;
  }
  
  if(M5.BtnC.wasPressed()){
    if(displayState){
      displayState = false;
      M5.Lcd.setBrightness(1);
    }
    else{
      displayState = true;
      M5.Lcd.setBrightness(100);
    }
  }
}

void drawCurrentTime(){

  int8_t tempHour = hour();
  int16_t hourX = 180;

  int8_t tempSeconds = second();
  tempSeconds %= 2;

  int8_t tempMinute = minute();
  bool decimalMinute = false;

  if(tempMinute<10)
    decimalMinute = true;


  if(tempHour<10)
    hourX = 202;

  
  M5.Lcd.setTextColor(WHITE, BLACK);
  M5.Lcd.setCursor(hourX, 94);            
  M5.Lcd.setTextSize(4 );
  M5.Lcd.print(tempHour);

  if(tempSeconds == 0)
    M5.Lcd.print(':');
  else
    M5.Lcd.print(' ');

  if(decimalMinute)
    M5.Lcd.print("0");

  M5.Lcd.print(tempMinute);   

  M5.Lcd.setCursor(182, 135);
  M5.Lcd.setTextSize(2);
  
  M5.Lcd.print(year());
  M5.Lcd.print('/');
  M5.Lcd.print(month());
  M5.Lcd.print('/');
  M5.Lcd.print(day());

  
}

void drawPLCWord(){
  M5.Lcd.setCursor(15, 99);
  M5.Lcd.setTextColor(YELLOW, BLACK);
  M5.Lcd.setTextSize(6);
  M5.Lcd.print("PLC");
}

void drawHMIWord(){
  M5.Lcd.setCursor(15, 99);
  M5.Lcd.setTextColor(ORANGE, BLACK);
  M5.Lcd.setTextSize(6);
  M5.Lcd.print("HMI");
}

void clearScreen(){
  M5.Lcd.fillScreen(0);
}

void drawBackground(){
  M5.Lcd.fillRect(0,0,320,80, GRAY_BACK);

  M5.Lcd.fillRect(0,158,320,80, GRAY_BACK);
}


void digitalPinO(bool state, int selectedPin){

  int16_t tempX = 4;
  int16_t tempY = 4;

  int16_t labelX = 30;
  int16_t labelY = 65;

  uint16_t tempColor = DARK_RED;
  char* tempPin = namePinO0;
  
  if(state)                        //(4,4) (W 72 - H 54)
      tempColor = LIGHT_GREEN;

  switch(selectedPin){
    case 1:
            tempX += PINO1X;
            labelX += PINO1X;
            tempPin = namePinO1;
    break;
    case 2:
            tempX += PINO2X;
            labelX += PINO2X;
            tempPin = namePinO2;
    break;
    case 3:
            tempX += PINO3X;
            labelX += PINO3X;
            tempPin = namePinO3;
    break;
  }

 M5.Lcd.fillRoundRect(tempX, tempY, 72, 54, 4, tempColor);
 M5.Lcd.setCursor(labelX, labelY);
 M5.Lcd.setTextColor(WHITE, GRAY_BACK);
 M5.Lcd.setTextSize(1);
 M5.Lcd.print(tempPin);

}//END digitalPinI0

void digitalPinI(bool state, int selectedPin){

  int16_t tempX = 4 + PINI0X;
  int16_t tempY = 4 + PINI0Y;

  int16_t labelX = 30 + PINI0X;
  int16_t labelY = 65 + PINI0Y;

  uint16_t tempColor = DARK_RED;
  char* tempPin = namePinI0;
  
  if(state)                        //(4,4) (W 72 - H 54)
      tempColor = LIGHT_GREEN;

  switch(selectedPin){
    case 1:
            tempX = 4 + PINI1X;
            tempY = 4 + PINI1Y;
            
            labelX = 30 + PINI1X;
            labelY = 65 + PINI1Y;
            
            tempPin = namePinI1;
    break;
    case 2:
            tempX = 4 + PINI2X;
            tempY = 4 + PINI2Y;
            
            labelX = 30 + PINI2X;
            labelY = 65 + PINI2Y;
            
            tempPin = namePinI2;
    break;
  }

 M5.Lcd.fillRoundRect(tempX, tempY, 72, 54, 4, tempColor);
 M5.Lcd.setCursor(labelX, labelY);
 M5.Lcd.setTextColor(WHITE, GRAY_BACK);
 M5.Lcd.setTextSize(1);
 M5.Lcd.print(tempPin);

}//END digitalPinI0


void analogPinO(int value, int selectedPin, bool servo){
  int16_t tempX = 4;
  int16_t tempY = 4;

  int16_t labelX = 30;
  int16_t labelY = 65;

  int16_t digitX = 35;

  uint16_t tempColor = RED;

  char* tempPin = namePinO0;

  if(servo)
    tempColor = BLUE_C;
  
  if((value > 9) && (value < 100))
    digitX = 29;
  else if((value > 99) && (value < 1000))  
    digitX = 23;
  else if(value > 999)  
    digitX = 17;

  switch(selectedPin){
    case 1:
            tempX += PINO1X;
            labelX += PINO1X;
            digitX += PINO1X;
            tempPin = namePinO1;
    break;
    case 2:
            tempX += PINO2X;
            labelX += PINO2X;
            digitX += PINO2X;
            tempPin = namePinO2;
    break;
    case 3:
            tempX += PINO3X;
            labelX += PINO3X;
            digitX += PINO3X;
            tempPin = namePinO3;
    break;
  }


  M5.Lcd.fillRoundRect(tempX, tempY, 72, 54, 4, BLACK);
  M5.Lcd.setCursor(labelX, labelY);
  M5.Lcd.setTextColor(WHITE, GRAY_BACK);
  M5.Lcd.setTextSize(1);
  M5.Lcd.print(tempPin);

  M5.Lcd.setCursor(digitX, 24);            //4 DIGITS (17) 3 (23) 2 (29) 1(35) 
  M5.Lcd.setTextColor(tempColor, BLACK);
  M5.Lcd.setTextSize(2);
  M5.Lcd.print(value);
}

void analogPinI(int value, int selectedPin){
  int16_t tempX = 4 + PINI0X;
  int16_t tempY = 4 + PINI0Y;

  int16_t labelX = 30 + PINI0X;
  int16_t labelY = 65 + PINI0Y;

  int16_t xOffset = 35;

  int16_t digitX = 35;
  int16_t digitY = 24 + PINI0Y;
  
  char* tempPin = namePinI0;
  
  if((value > 9) && (value < 100))
    digitX = 29;
  else if((value > 99) && (value < 1000))  
    digitX = 23;
  else if(value > 999)  
    digitX = 17;
      

  switch(selectedPin){
    case 0:
            digitX += PINI0X;
    break;
    case 1:
            tempX = 4 + PINI1X;
            tempY = 4 + PINI1Y;
            
            labelX = 30 + PINI1X;
            labelY = 65 + PINI1Y;

            digitX += PINI1X;
            digitY = 24 + PINI1Y;
            
            tempPin = namePinI1;
    break;
    case 2:
            tempX = 4 + PINI2X;
            tempY = 4 + PINI2Y;
            
            labelX = 30 + PINI2X;
            labelY = 65 + PINI2Y;

            digitX += PINI2X;
            digitY = 24 + PINI2Y;
            
            tempPin = namePinI2;
    break;
  }

  M5.Lcd.fillRoundRect(tempX, tempY, 72, 54, 4, BLACK);
  M5.Lcd.setCursor(labelX, labelY);
  M5.Lcd.setTextColor(WHITE, GRAY_BACK);
  M5.Lcd.setTextSize(1);
  M5.Lcd.print(tempPin);

  M5.Lcd.setCursor(digitX, digitY);            //4 DIGITS (17) 3 (23) 2 (29) 1(35) 
  M5.Lcd.setTextColor(RED, BLACK);
  M5.Lcd.setTextSize(2);
  M5.Lcd.print(value);
  
}

void showProgramming(){
 
  M5.Lcd.fillScreen(0);
  M5.Lcd.setCursor(4,1);
  M5.Lcd.print("UPLOADING DESIGN");
  currentTimeDisplay = 0;
  startTimeDisplay = millis();
  while(programming){
    currentTimeDisplay += millis();
    if(currentTimeDisplay >= 600){
      M5.Lcd.print(".");
      currentTimeDisplay = 0;
    }
    delay(200);
  }//END while
  currentTimeDisplay = 0;

}

void preDraw(){

  int tempValue;
  
  for(int i=0; i<3; i++){
    if(analogIEnabled[i]){
      preAnalogInputs[i] = analogInputs[i]; 
      analogPinI(analogInputs[i], i);  
       
    }
    else{
      preIState[i] = iState[i];
      digitalPinI(iState[i], i);
           
    }
  }//END for


  for(int i=0; i<4; i++){
    if((pwmOutput[i]) || (servoOutput[i])){
      bool isServo = true;
      if(pwmOutput[i])
        isServo = false;
        
      preAnalogOutputs[i] = analogOutputs[i];
      analogPinO(analogOutputs[i], i, isServo);
   
    }
    else{
        preOState[i] = oState[i];
        digitalPinO(oState[i],i);
      
    }
  }//END for

  /*
  for(int i=0; i<3; i++){
    digitalPinI(false,i);
  }

  for(int i=0; i<4; i++){
    digitalPinO(false, i);
  }*/
}

void stopProcess(){
  while(true){
    delay(1000);
  }
}

void showResetM5(){
  M5.Lcd.fillScreen(YELLOW);
  M5.Lcd.setCursor(4,1);
  M5.Lcd.setTextSize(4);
  M5.Lcd.setTextColor(BLACK, YELLOW);
  M5.Lcd.println(" UPLOAD");
  M5.Lcd.println(" COMPLETED");
  M5.Lcd.setTextColor(RED, YELLOW);
  M5.Lcd.println("");
  M5.Lcd.println(" RESET");
  M5.Lcd.println(" YOUR");
  M5.Lcd.println(" M5STACK");
  

  while(true){
    delay(1000);
  }
  
}




